# RCAIDE/Framework/Analyses/Mission/Segments/Conditions/Conditions.py
# 
# 
# Created:  Jul 2023, M. Clarke
 
# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
# RCAIDE imports 
from RCAIDE.Framework.Core                    import Data 

# python imports 
import numpy as np 
# ----------------------------------------------------------------------------------------------------------------------
#  Conditions
# ----------------------------------------------------------------------------------------------------------------------
class Conditions(Data):
    """ Conditions are the magic Data that contains the information about the vehicle in flight.
        At this point none of the information really exists. What is here are the methods that allow a mission
        to collect the information.
    
        Assumptions:
        None
        
        Source:
        None   
    """ 

    _size = 1
    
    def ones_row(self,cols):
        """ returns a row vector of ones with given number of columns 
        
            Assumptions:
            None
    
            Source:
            N/A
    
            Inputs:
            cols   [in]
    
            Outputs:
            Vector
    
            Properties Used:
            None
        """     
        return np.ones([self._size,cols])
    
    def ones_row_m1(self,cols):
        """ returns an N-1 row vector of ones with given number of columns
        
            Assumptions:
            None
    
            Source:
            N/A
    
            Inputs:
            cols   [in]
    
            Outputs:
            Vector
    
            Properties Used:
            None
        """ 
        return expanded_array(cols, 1)
    
    def ones_row_m2(self,cols):
        """ returns an N-2 row vector of ones with given number of columns
        
            Assumptions:
            None
    
            Source:
            N/A
    
            Inputs:
            cols   [int]
    
            Outputs:
            Vector
    
            Properties Used:
            None
        """ 
        return expanded_array(cols, 2)
    
    
    def expand_rows(self,rows,override=False):
        """ Makes a 1-D array the right size. Often used after a mission is initialized to size out the vectors to the
            right size. Will not overwrite an array if it already exists, unless override is True.
        
            Assumptions:
            None
    
            Source:
            N/A
    
            Inputs:
            rows     [int]
            override [boolean]
    
            Outputs:
            None
    
            Properties Used:
            None
        """           
        
        # store
        self._size = rows
        
        # recursively initialize condition and unknown arrays 
        # to have given row length
        
        for k,v in self.items():
            try:
                rank = v.ndim
            except:
                rank = 0
            # recursion
            if isinstance(v,Conditions):
                v.expand_rows(rows,override=override)
            elif isinstance(v,expanded_array):
                self[k] = v.resize(rows)
            # need arrays here
            elif rank == 2:
                #Check if it's already expanded
                if v.shape[0]<=1 or override:
                    self[k] = np.resize(v,[rows,v.shape[1]])
        
        return
                
class expanded_array(Data):
    """ This is an array that will expand later when the mission is initialized. It is called specifically by conditions
    
        Assumptions:
        None
        
        Source:
        None   
    """ 

    _size = 1  
        
    def __init__(self, cols, adjustment):
        """ Initialization that sets expansion later
        
            Assumptions:
            None
        
            Source:
            N/A
        
            Inputs:
            self
            cols       - columns                          [int]
            adjustment - how much smaller                 [int]
        
            Outputs:
            N/A
            
            Properties Used:
            N/A
        """          
        
        self._adjustment = adjustment
        self._cols       = cols
        self._array      = np.array([[1]])
        
        
    def resize(self,rows):
        """ This function actually completes the resizing. After this it's no longer an expanded array. That way it
            doesn't propogate virally. That means if one wishes to resize later the conditions need to be reset.
        
            Assumptions:
            None
        
            Source:
            N/A
        
            Inputs:
            self
            rows       - rows                             [int]
            v          - values (really self)             [int]
        
            Outputs:
            np.array   - properly sized                   [array]
            
            Properties Used:
            N/A
        """   
        # unpack
        adjustment = self._adjustment
        
        # pack
        self._size = rows
        value      = self._array
        
        return np.resize(value,[rows-adjustment,value.shape[1]])
    
    def __call__(self):
        """ This returns the value and shape of the array as is
        
            Assumptions:
            None
        
            Source:
            N/A
        
            Inputs:
            self

            Outputs:
            np.array   - properly sized                   [array]
            
            Properties Used:
            N/A
        """           
        
        return self._array
    
    def __mul__(self,other):
        """ Performs multiplication and returns self
        
            Assumptions:
            None
        
            Source:
            N/A
        
            Inputs:
            self
            other      - something can be multiplied      [float]

            Outputs:
            self
            
            Properties Used:
            N/A
        """          
        
        self._array = np.resize(other,[1,1])
        
        return self

    def __rmul__(self,other):
        """ Performs multiplication and returns self
        
            Assumptions:
            None
        
            Source:
            N/A
        
            Inputs:
            self
            other      - something can be multiplied      [float]

            Outputs:
            self
            
            Properties Used:
            N/A
        """                 
        
        self._array = np.resize(other,[1,1])
        
        return self    
        
    


# RCAIDE/Library/Plots/Weights/plot_weight_breakdown.py
# 
# 
# Created:  Jul 2023, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
import RCAIDE 
import numpy as np     
import plotly.express as px 
import pandas as pd

# ----------------------------------------------------------------------------------------------------------------------
#  PLOTS
# ----------------------------------------------------------------------------------------------------------------------   
def plot_weight_breakdown(vehicle,
                            save_figure    = False,
                            show_figure    = True, 
                            show_legend    = True,
                            SI_Units       = True,
                            save_filename  = "Weight_Breakdown",
                            aircraft_name  = None,
                            file_type      = ".png",
                            width          = 10, height = 7.2): 
  

    """This plots the weight breakdown of an evtol aircraft

    Assumptions:
    None

    Source:
    None

    Inputs:
    vehicle

    Outputs:
    Plots

    Properties Used:
    N/A
    """

    breakdown =  vehicle.weight_breakdown     
    
    level_1 = []
    level_2 = []
    level_3 = []
    values  = []
      
    for tag ,  item in  breakdown.items():
        if tag == 'zero_fuel_weight' or   tag == 'max_takeoff':
            pass
        else:
            if type(item)  == RCAIDE.Framework.Core.Data: 
                for  sub_tag  , sub_item in item.items():
                    if type(sub_item) == RCAIDE.Framework.Core.Data: 
                        for sub_sub_tag ,  sub_sub_item in sub_item.items():
                            if sub_sub_tag != 'total':
                                level_1.append(tag.replace("_", " "))
                                level_2.append(sub_tag.replace("_", " "))
                                level_3.append(sub_sub_tag.replace("_", " "))
                                values.append(sub_sub_item )
                    elif sub_tag != 'total':
                        level_1.append(tag.replace("_", " "))
                        level_2.append(sub_tag.replace("_", " "))
                        level_3.append(np.nan)
                        values.append(sub_item) 
            elif tag == 'fuel':
                level_1.append(tag.replace("_", " "))
                level_2.append(np.nan)
                level_3.append(np.nan)
                values.append(sub_item) 
                  
    df = pd.DataFrame(
        dict(level_1=level_1, level_2=level_2, level_3=level_3, values=values)
    ) 
    fig = px.sunburst(df,
                      path=['level_1', 'level_2', 'level_3'], 
                      values='values',  
                      color_discrete_sequence=px.colors.qualitative.G10)
    
    # Add a dummy inner layer for the hole
    fig.update_traces(
        textfont=dict(size=20), 
        insidetextorientation='horizontal', 
        textinfo='label+percent entry', 
        marker=dict(colors=['rgba(0,0,0,0)'] + px.colors.qualitative.G10)
    )
    fig.update_layout( 
    uniformtext=dict(minsize=12, mode='hide'),
    )
 
    if save_figure:
        fig.write_image(save_filename + file_type)
    
    if show_figure:
        fig.write_html( save_filename + '.html', auto_open=True)        
    return  fig 



# RCAIDE/Methods/__init__.py
# 

""" RCAIDE Package Setup
"""

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

from . import Aerodynamics 
from . import Emissions
from . import Energy
from . import Gas_Dynamics
from . import Geometry
from . import Noise
from . import Performance 
from . import Propulsors
from . import Stability
from . import Utilities
from . import Weights
from . import Geodesics

from .skip import skip




# RCAIDE/Library/Plots/Geometry/plot_3d_nacelle.py
# 
# 
# Created:  Jul 2023, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------  
import RCAIDE
from RCAIDE.Framework.Core import Data 
from RCAIDE.Library.Plots.Geometry.Common.contour_surface_slice import contour_surface_slice
from RCAIDE.Library.Methods.Geometry.Airfoil import import_airfoil_geometry
from RCAIDE.Library.Methods.Geometry.Airfoil import compute_naca_4series  
import numpy as np  

# ----------------------------------------------------------------------------------------------------------------------
#  PLOTS
# ----------------------------------------------------------------------------------------------------------------------   
def plot_3d_nacelle(plot_data,nacelle,tessellation = 24,number_of_airfoil_points = 21,color_map= 'darkmint'):
    """ This plots a 3D surface of a nacelle  

    Assumptions:
    None

    Source:
    None 
    
    Inputs:
    axes                       - plotting axes 
    nacelle                    - RCAIDE nacelle data structure
    tessellation               - azimuthal discretization of lofted body 
    number_of_airfoil_points   - discretization of airfoil geometry 
    color_map                  - face color of nacelle  

    Properties Used:
    N/A
    """
    
    if type(nacelle) == RCAIDE.Library.Components.Nacelles.Stack_Nacelle: 
        G = generate_3d_stack_nacelle_points(nacelle,tessellation,number_of_airfoil_points)
    elif type(nacelle) == RCAIDE.Library.Components.Nacelles.Body_of_Revolution_Nacelle: 
        G = generate_3d_BOR_nacelle_points(nacelle,tessellation,number_of_airfoil_points)
    else:
        G= generate_3d_basic_nacelle_points(nacelle,tessellation,number_of_airfoil_points)
        
               
    num_nac_segs = len(G.PTS[:,0,0])
    tesselation  = len(G.PTS[0,:,0]) 
    for i_seg in range(num_nac_segs-1):
        for i_tes in range(tesselation-1):
            X = np.array([[G.PTS[i_seg  ,i_tes  ,0],G.PTS[i_seg+1,i_tes  ,0]],
                 [G.PTS[i_seg  ,i_tes+1,0],G.PTS[i_seg+1,i_tes+1,0]]])
            Y = np.array([[G.PTS[i_seg  ,i_tes  ,1],G.PTS[i_seg+1,i_tes  ,1]],
                 [G.PTS[i_seg  ,i_tes+1,1],G.PTS[i_seg+1,i_tes+1,1]]])
            Z = np.array([[G.PTS[i_seg  ,i_tes  ,2],G.PTS[i_seg+1,i_tes  ,2]],
                 [G.PTS[i_seg  ,i_tes+1,2],G.PTS[i_seg+1,i_tes+1,2]]])
             
            values = np.zeros_like(X) 
            verts = contour_surface_slice(X, Y, Z ,values,color_map)
            plot_data.append(verts)    

    return plot_data
 
def generate_3d_stack_nacelle_points(nac,tessellation = 24 ,number_of_airfoil_points = 21):
    """ This generates the coordinate points on the surface of the stack nacelle

    Assumptions:
    None

    Source:
    None

    Inputs:
    nac                        - Nacelle data structure 
    tessellation               - azimuthal discretization of lofted body 
    number_of_airfoil_points   - discretization of airfoil geometry 
    
    Properties Used:
    N/A 
    """ 
    
    num_nac_segs = len(nac.segments.keys())   
    theta        = np.linspace(0,2*np.pi,tessellation)  
 
    nac_pts = np.zeros((num_nac_segs,tessellation,3))  
    for i_seg, segment in enumerate(nac.segments): 
        a = segment.width/2
        b = segment.height/2
        n = segment.curvature
        theta                   = np.linspace(0,2*np.pi,tessellation) 
        section_points          = np.zeros((24, 3, 1))
        section_points[:, 1, 0] = (abs((np.cos(theta)))**(2/n))*a * ((np.cos(theta)>0)*1 - (np.cos(theta)<0)*1) 
        section_points[:, 2, 0] = (abs((np.sin(theta)))**(2/n))*b * ((np.sin(theta)>0)*1 - (np.sin(theta)<0)*1)  
        
        x_rotation = np.zeros((tessellation, 3, 3))
        x_rotation[:,0,0] = 1
        x_rotation[:,1,1] = np.cos(segment.orientation_euler_angles[0])
        x_rotation[:,1,2] = -np.sin(segment.orientation_euler_angles[0])
        x_rotation[:,2,1] = np.sin(segment.orientation_euler_angles[0])
        x_rotation[:,2,2] = np.cos(segment.orientation_euler_angles[0])
        
        y_rotation = np.zeros((tessellation, 3, 3))
        y_rotation[:,0,0] = np.cos(segment.orientation_euler_angles[1])
        y_rotation[:,0,2] = np.sin(segment.orientation_euler_angles[1])
        y_rotation[:,1,1] = 1
        y_rotation[:,2,0] = -np.sin(segment.orientation_euler_angles[1])
        y_rotation[:,2,2] = np.cos(segment.orientation_euler_angles[1]) 

        z_rotation = np.zeros((tessellation, 3, 3))
        z_rotation[:,0,0] = np.cos(segment.orientation_euler_angles[2])
        z_rotation[:,0,1] = -np.sin(segment.orientation_euler_angles[2])
        z_rotation[:,1,0] = np.sin(segment.orientation_euler_angles[2])
        z_rotation[:,1,1] = np.cos(segment.orientation_euler_angles[2])
        z_rotation[:,2,2] = 1
        
        rotated_pts = np.matmul(z_rotation,np.matmul(y_rotation,np.matmul(x_rotation,section_points))) 

        nac_pts[i_seg] = rotated_pts[:,:,0]   

        nac_pts[i_seg,:,0] += segment.percent_x_location*nac.length     
        nac_pts[i_seg,:,1] += segment.percent_y_location*nac.length 
        nac_pts[i_seg,:,2] += segment.percent_z_location*nac.length          
            
    # rotation about y to orient propeller/rotor to thrust angle
    rot_trans =  nac.nac_vel_to_body()
    rot_trans =  np.repeat( np.repeat(rot_trans[ np.newaxis,:,: ],tessellation,axis=0)[ np.newaxis,:,:,: ],num_nac_segs,axis=0)    
    
    NAC_PTS  =  np.matmul(rot_trans,nac_pts[...,None]).squeeze()  
     
    # translate to body 
    NAC_PTS[:,:,0] = NAC_PTS[:,:,0] + nac.origin[0][0]
    NAC_PTS[:,:,1] = NAC_PTS[:,:,1] + nac.origin[0][1]
    NAC_PTS[:,:,2] = NAC_PTS[:,:,2] + nac.origin[0][2]
     
    G = Data()         
    G.PTS  = NAC_PTS 
    return G     


 
def generate_3d_BOR_nacelle_points(nac,tessellation = 24 ,number_of_airfoil_points = 21):
    """ This generates the coordinate points on the surface of the body of revolution nacelle

    Assumptions:
    None

    Source:
    None

    Inputs:
    nac                        - Nacelle data structure 
    tessellation               - azimuthal discretization of lofted body 
    number_of_airfoil_points   - discretization of airfoil geometry 
    
    Properties Used:
    N/A 
    """ 
     
    theta        = np.linspace(0,2*np.pi,tessellation)  
    num_nac_segs = int(np.ceil(number_of_airfoil_points/2))
    nac_pts      = np.zeros((num_nac_segs,tessellation,3)) 
    naf          = nac.airfoil 
    
    if type(naf) == RCAIDE.Library.Components.Airfoils.NACA_4_Series_Airfoil: 
        a_geo        = compute_naca_4series(naf.NACA_4_Series_code,num_nac_segs)
        xpts         = np.repeat(np.atleast_2d(a_geo.x_coordinates).T,tessellation,axis = 1)*nac.length
        zpts         = np.repeat(np.atleast_2d(a_geo.y_coordinates).T,tessellation,axis = 1)*nac.length  
    
    elif naf.coordinate_file != None: 
        a_geo        = import_airfoil_geometry(naf.coordinate_file,num_nac_segs)
        xpts         = np.repeat(np.atleast_2d(np.take(a_geo.x_coordinates,axis=0)).T,tessellation,axis = 1)*nac.length
        zpts         = np.repeat(np.atleast_2d(np.take(a_geo.y_coordinates,axis=0)).T,tessellation,axis = 1)*nac.length 

    if nac.flow_through: 
        zpts = zpts + nac.diameter/2  
            
    # create geometry 
    theta_2d = np.repeat(np.atleast_2d(theta),num_nac_segs,axis =0) 
    nac_pts[:,:,0] =  xpts
    nac_pts[:,:,1] =  zpts*np.cos(theta_2d)
    nac_pts[:,:,2] =  zpts*np.sin(theta_2d)   
            
    # rotation about y to orient propeller/rotor to thrust angle
    rot_trans =  nac.nac_vel_to_body()
    rot_trans =  np.repeat( np.repeat(rot_trans[ np.newaxis,:,: ],tessellation,axis=0)[ np.newaxis,:,:,: ],num_nac_segs,axis=0)    
    
    NAC_PTS  =  np.matmul(rot_trans,nac_pts[...,None]).squeeze()  
     
    # translate to body 
    NAC_PTS[:,:,0] = NAC_PTS[:,:,0] + nac.origin[0][0]
    NAC_PTS[:,:,1] = NAC_PTS[:,:,1] + nac.origin[0][1]
    NAC_PTS[:,:,2] = NAC_PTS[:,:,2] + nac.origin[0][2]
    
    G = Data()         
    G.PTS  = NAC_PTS 
    return G     

def generate_3d_basic_nacelle_points(nac,tessellation,number_of_airfoil_points):
    """ This generates the coordinate points on the surface of the nacelle

    Assumptions:
    None

    Source:
    None

    Inputs:
    nac                        - Nacelle data structure 
    tessellation               - azimuthal discretization of lofted body 
    number_of_airfoil_points   - discretization of airfoil geometry 
    
    Properties Used:
    N/A 
    """ 
      

    num_nac_segs = 10
    nac_pts      = np.zeros((num_nac_segs,tessellation,3)) 
    theta        = np.linspace(0,2*np.pi,tessellation)  
          
    # if no airfoil defined, use super ellipse as default
    a    =  nac.length/2 
    b    =  nac.diameter/2  
    xpts =  np.repeat(np.atleast_2d(np.linspace(-a,a,num_nac_segs)).T,tessellation,axis = 1) 
    zpts = np.sqrt((b**2)*(1 - (xpts**2)/(a**2) ))  
    xpts = (xpts+a)  

    if nac.flow_through: 
        zpts = zpts + nac.inlet_diameter/2  
            
    # create geometry 
    theta_2d = np.repeat(np.atleast_2d(theta),num_nac_segs,axis =0) 
    nac_pts[:,:,0] =  xpts
    nac_pts[:,:,1] =  zpts*np.cos(theta_2d)
    nac_pts[:,:,2] =  zpts*np.sin(theta_2d)  
                 
    # rotation about y to orient propeller/rotor to thrust angle
    rot_trans =  nac.nac_vel_to_body()
    rot_trans =  np.repeat( np.repeat(rot_trans[ np.newaxis,:,: ],tessellation,axis=0)[ np.newaxis,:,:,: ],num_nac_segs,axis=0)    
    
    NAC_PTS  =  np.matmul(rot_trans,nac_pts[...,None]).squeeze()  
     
    # translate to body 
    NAC_PTS[:,:,0] = NAC_PTS[:,:,0] + nac.origin[0][0]
    NAC_PTS[:,:,1] = NAC_PTS[:,:,1] + nac.origin[0][1]
    NAC_PTS[:,:,2] = NAC_PTS[:,:,2] + nac.origin[0][2]
     
    G = Data()         
    G.PTS  = NAC_PTS 
    return G     


## @ingroup Optimization-Package_Setups
# scipy_setup.py
# 
# Created:  Aug 2015, E. Botero 
# Modified: Feb 2017, M. Vegh
#           Mar 2020, E. Botero
#           Jul 2020, M. Clarke
#           May 2021, E. Botero 

# ----------------------------------------------------------------------
#  Imports
# ----------------------------------------------------------------------

# rcaide imports
import numpy as np
import scipy as sp
from RCAIDE.Framework.Optimization.Packages.particle_swarm import particle_swarm_optimization 
from scipy.optimize import NonlinearConstraint
from RCAIDE.Framework.Optimization.Common import helper_functions as help_fun

# ----------------------------------------------------------------------
#  Something that should become a class at some point
# ----------------------------------------------------------------------

## @ingroup Optimization-Package_Setups
def SciPy_Solve(problem,solver='SLSQP', sense_step = 1.4901161193847656e-08, iter =200, tolerance = 1e-6, pop_size =  10 , prob_seed = None ):  
    """ This converts your RCAIDE Nexus problem into a SciPy optimization problem and solves it
        SciPy has many algorithms, they can be switched out by using the solver input. 

        Assumptions:
        1.4901161193847656e-08 is SLSQP default FD step in scipy

        Source:
        N/A

        Inputs:
        problem                   [nexus()]
        solver                    [str]
        sense_step                [float]

        Outputs:
        outputs                   [list]

        Properties Used:
        None
    """
    
    inp = problem.optimization_problem.inputs
    obj = problem.optimization_problem.objective
    con = problem.optimization_problem.constraints
    
    # Have the optimizer call the wrapper
    wrapper = lambda x:SciPy_Problem(problem,x)    
    
    # Set inputsq
    nam  = inp[:,0] # Names
    ini  = inp[:,1] # Initials
    bndl = inp[:,2] # Bounds
    bndu = inp[:,3] # Bounds
    scl  = inp[:,4] # Scale
    
    x   = ini/scl
    bnds = np.zeros((len(inp),2))
    lb   = np.zeros(len(inp))
    ub   = np.zeros(len(inp))
    de_bnds = []    
    
    for ii in range(0,len(inp)):
        # Scaled bounds
        bnds[ii] = (bndl[ii]/scl[ii]),(bndu[ii]/scl[ii])  
        lb[ii]   = bndl[ii]/scl[ii]
        ub[ii]   = bndu[ii]/scl[ii]
        de_bnds.append((bndl[ii]/scl[ii],bndu[ii]/scl[ii]))  
     
    # Finalize problem statement and run
    if solver=='SLSQP':
        outputs = sp.optimize.fmin_slsqp(wrapper,x,f_eqcons=problem.equality_constraint,f_ieqcons=problem.inequality_constraint,bounds=bnds,\
                                         iter=iter, epsilon = sense_step, acc  = tolerance)
    elif solver == 'differential_evolution':
        # Define constraints as a tuple of nonlinear constraints 
        scaled_constraints = []
        aliases            = problem.optimization_problem.aliases
        for ii in range(0,len(con)):
            de_constraint  = con[[ii]] 
            def fun(x):
                problem.evaluate(x)
                constraint_val = help_fun.get_values(problem,de_constraint,aliases)
                return np.atleast_1d(constraint_val)
            
            bound  = help_fun.scale_const_bnds(con)
            if con[ii][1]=='=':
                print('Nonlinear constraints for scipy differential evoultion optimization has '
                      'the general inequality form. Consider rewriting equality constraint as two '
                      'separate inequality constraints')
            
            if con[ii][1]=='>':
                nlc = NonlinearConstraint(fun,bound[ii], np.inf) 
                
            elif con[ii][1]=='<':
                nlc = NonlinearConstraint(fun, -np.inf,bound[ii])
                
            scaled_constraints.append(nlc) 
            
        diff_evo_cons = tuple(scaled_constraints)    
        
        outputs = sp.optimize.differential_evolution(wrapper, bounds= de_bnds, strategy='best1bin', maxiter=1000, popsize = pop_size, \
                                                     tol=0.01, mutation=(0.5, 1), recombination=0.7, seed=prob_seed, callback=None,\
                                                     disp=False, polish=True, init='latinhypercube', atol=0, updating='immediate',\
                                                     workers=1,constraints=diff_evo_cons)
        
    elif solver == 'particle_swarm_optimization':
        outputs = particle_swarm_optimization(wrapper, lb, ub, f_ieqcons=problem.inequality_constraint, kwargs={}, swarmsize=pop_size ,\
                                              omega=0.5, phip=0.5, phig=0.5, maxiter=1000, minstep=1e-4, minfunc=1e-4, debug=False)    
    else:
        outputs = sp.optimize.minimize(wrapper,x,method=solver)
    
    return outputs

## @ingroup Optimization-Package_Setups
def SciPy_Problem(problem,x):
    """ This wrapper runs the RCAIDE problem and is called by the Scipy solver.
        Prints the inputs (x) as well as the objective value

        Assumptions:
        None

        Source:
        N/A

        Inputs:
        problem   [nexus()]
        x         [array]

        Outputs:
        obj       [float]

        Properties Used:
        None
    """      
    
    print('Inputs')
    print(x)        
    obj   = problem.objective(x)
    print('Obj')
    print(obj)

    
    return obj




# RCAIDE/Framework/Optimization/Common/helper_functions.py 

# ----------------------------------------------------------------------------------------------------------------- 
#  IMPORT
# --- ------------------------------------------------------------------------------------------------------------- 
  
import numpy as np

# ----------------------------------------------------------------------------------------------------------------- 
#  Set Values
# -----------------------------------------------------------------------------------------------------------------  
def set_values(dictionary,input_dictionary,converted_values,aliases):
    """ This method regresses through a dictionary to set the required values.
        dictionary is the base class that will be modified, input_dictionary is
        the set of inputs to be used, converted_values are values to be set in the
        base dictionary, and finally the aliases which are where in the dictionary
        the names link to

    Assumptions:
    N/A

    Source:
    N/A

    Inputs:
    dictionary       [Data()]
    input_dictionary [Data()]
    converted_values [Data()]
    aliases          [list of str]

    Outputs:
    None

    Properties Used:
    N/A
    """      
    
    provided_names = input_dictionary[:,0]
        
    # Correspond aliases to inputs
    pointer = []
    for ii in range(0,len(provided_names)):
        for jj in range(0,len(aliases)):
            if provided_names[ii] == aliases[jj][0]:
                pointer.append(aliases[jj][1])

    for ii in range(0,len(pointer)):
        pointers = pointer[ii][:]
        if isinstance(pointers,str):
            if '*' in pointers:
                newstrings = find_a_star(dictionary,pointer[ii])
                for jj in range(0,len(newstrings)):
                    localstrs = newstrings[jj].split('.')
                    correctname = '.'.join(localstrs[0:])
                    dictionary.deep_set(correctname,converted_values[ii])
                    
            elif '*' not in pointers:
                localstrs = pointers.split('.')
                correctname = '.'.join(localstrs[0:])
                dictionary.deep_set(correctname,converted_values[ii])              
        else:
            for zz in range(1,len(pointers)+1):
                if '*' in pointers[zz-1]:
                    newstrings = find_a_star(dictionary,pointer[ii][zz-1])
                    for jj in range(0,len(newstrings)):
                        localstrs = newstrings[jj].split('.')
                        correctname = '.'.join(localstrs[0:])
                        dictionary.deep_set(correctname,converted_values[ii])
                        
                elif '*' not in pointers[zz-1]:
                    localstrs = pointers[zz-1].split('.')
                    correctname = '.'.join(localstrs[0:])
                    dictionary.deep_set(correctname,converted_values[ii])            
            
    return dictionary
        

def find_a_star(dictionary,string):
    """ Searches through a dictionary looking for an *

    Assumptions:
    There may or may not be an asterisk

    Source:
    N/A

    Inputs:
    dictionary       [Data()]
    input_dictionary [Data()]
    converted_values [Data()]
    aliases          [list of str]

    Outputs:
    newstrings       [list of str]

    Properties Used:
    N/A
    """
    splitstring = string.split('.')
    for ii in range(0,len(splitstring)):
        if '*' in splitstring[ii]:
            if ii==0:
                newkeys = dictionary.keys()
            elif ii !=0:
                strtoeval = 'dictionary.'+'.'.join(splitstring[0:ii])+'.keys()'
                newkeys = list(eval(strtoeval))
            lastindex   = ii
            
    newstrings = []
    for ii in range(0,len(newkeys)):
        newstrings.append('.'.join(splitstring[0:lastindex])+'.'+newkeys[ii]+'.'+'.'.join(splitstring[lastindex+1:]))
        
    return newstrings


def scale_input_values(inputs,x):
    """ Scales the values according to the a provided scale

    Assumptions:
    

    Source:
    N/A

    Inputs:
    x                [array]         
    inputs           [list]

    Outputs:
    inputs           [list]

    Properties Used:
    N/A
    """    
    
    provided_scale = inputs[:,-2]
    inputs[:,1] =  x*provided_scale
    
    return inputs

def limit_input_values(inputs):
    """ Ensures that the inputs are between the bounds

    Assumptions:
    N/A

    Source:
    N/A

    Inputs:
    x                [array]         
    inputs           [list]

    Outputs:
    inputs           [list]

    Properties Used:
    N/A
    """      
    
    provided_values = inputs[:,1]
    lower_bounds    = inputs[:,2]
    upper_bounds    = inputs[:,3]
    
    # Fix if the input is too low
    provided_values[provided_values<lower_bounds] = lower_bounds[provided_values<lower_bounds] 
    
    # Fix if the input is too high
    provided_values[provided_values>upper_bounds] = upper_bounds[provided_values>upper_bounds]

    
    return inputs
    


def convert_values(inputs): 
    """ Converts an inputs from an optimization into the right units

    Assumptions:
    Always multiply the units by 1!

    Source:
    N/A

    Inputs:
    inputs           [list]

    Outputs:
    converted_values [list of str]

    Properties Used:
    N/A
    """    
    
    provided_values  = inputs[:,1] 
    
    # Most important 2 lines of these functions
    provided_units   = inputs[:,-1]*1.0
    inputs[:,-1]     = provided_units
    
    converted_values = provided_values*provided_units
    
    return converted_values


# ----------------------------------------------------------------------        
#   Get
# ----------------------------------------------------------------------  


def get_values(dictionary,outputs,aliases):
    """ Retrieves values saved in a dictionary 

    Assumptions:
    N/A

    Source:
    N/A

    Inputs:
    dictionary       [Data()]
    outputs          [Data()]
    aliases          [list of str]

    Outputs:
    values           [float]

    Properties Used:
    N/A
    """     
    
    npoutputs   = np.array(outputs)
    output_names = npoutputs[:,0]
        
    # Correspond aliases to outputs
    pointer = []
    for ii in range(0,len(output_names)):
        for jj in range(0,len(aliases)):
            if output_names[ii] == aliases[jj][0]:
                pointer.append(aliases[jj][1])    
                
    values = np.zeros(len(outputs))
    for ii in range(0,len(outputs)):
        
        if isinstance(pointer[ii], str) and not ('*' in pointer[ii]) :
            splitstring = pointer[ii].split('.')
            
        else :
            raise TypeError("Pointers for Objectives and Constraints must be unique path (str), not list or contain asterisk")
            
        values[ii]  = eval('dictionary.'+'.'.join(splitstring[0:]))
    
    return values


def scale_obj_values(inputs,x):
    """ Rescales an objective based on Nexus inputs scale

    Assumptions:
    N/A

    Source:
    N/A

    Inputs:
    inputs          [Data()]
    x               [float]

    Outputs:
    scaled          [float]

    Properties Used:
    N/A
    """     
    
    provided_scale = inputs[:,1]
    provided_units = inputs[:,-1]*1.0
    inputs[:,-1]   = provided_units
    
    scaled =  x/(provided_scale*provided_units)
    
    return scaled


def scale_const_values(inputs,x):
    """ Rescales constraint values based on Nexus inputs scale

    Assumptions:
    N/A

    Source:
    N/A

    Inputs:
    inputs          [Data()]
    x               [array]

    Outputs:
    scaled          [array]

    Properties Used:
    N/A
    """        
    
    provided_scale = np.array(inputs[:,3],dtype = float)
    scaled =  x/provided_scale
    
    return scaled


def scale_const_bnds(inputs):
    """ Rescales constraint bounds based on Nexus inputs scale

    Assumptions:
    N/A

    Source:
    N/A

    Inputs:
    inputs           [Data()]

    Outputs:
    converted_values [array]

    Properties Used:
    N/A
    """     
    
    provided_bounds = np.array(inputs[:,2],dtype = float)
    
    # Most important 2 lines of these functions
    provided_units  = inputs[:,-1]*1.0
    inputs[:,-1]    = provided_units
    
    converted_values = provided_bounds*provided_units
    
    return converted_values


def unscale_const_values(inputs,x):
    """ Rescales values based on Nexus inputs scale

    Assumptions:
    N/A

    Source:
    N/A

    Inputs:
    inputs           [Data()]
    x                [array]

    Outputs:
    scaled           [array]

    Properties Used:
    N/A
    """     
    
    provided_units   = inputs[:,-1]*1.0
    provided_scale = np.array(inputs[:,-2],dtype = float)
    scaled =  x*provided_scale/provided_units
    
    return scaled



## @ingroup Methods-Aerodynamics-Common-Fidelity_Zero-Lift
# generate_wing_wake_grid.py
# 
# Created:   April 2021, R. Erhard
# Modified: 

# ----------------------------------------------------------------------
#  Imports
# ----------------------------------------------------------------------
import numpy as np 
from RCAIDE.Framework.Core import Data


def generate_wing_wake_grid(geometry, H, L, hf, x_plane, Nzo=20, Nzf=35, Nyo=20):
    """ Generates the grid points for evaluating the viscous wing wake in a downstream plane.
    Uses smaller grid near the wing to better capture boundary layer.
    
    Inputs: 
    geometry     RCAIDE vehicle data object
    H            Height of full grid, normalized by wing span
    L            Length of grid, normalized by wing span
    hf           Height of finer grid portion
    x_plane      Spanwise location of grid plane
    
    Nzo          Number of vertical grid points outside finer region
    Nzf          Number of vertical grid points inside finer region
    Nyo          Number of horizontal grid points outside of wing span
    """
    # unpack
    span      = geometry.wings.main_wing.spans.projected
    half_span = span/2
    VD        = geometry.vortex_distribution
    breaks    = VD.chordwise_breaks
    
    # grid bounds
    z_bot     = -H*half_span
    z_top     = H*half_span
    Nzo_half  = int(Nzo/2)
    Nyo_half  = int(Nyo/2)
       
    # generate vertical grid point locations
    z_outer_bot   = np.linspace(z_bot, -hf, Nzo_half)
    z_outer_top   = np.linspace(hf, z_top, Nzo_half)
    
    # use finer concentration of grid points near the wing
    z_inner_bot   = -hf*(np.flipud((1-np.cos(np.linspace(1e-6,1,Nzf)*np.pi/2))))
    z_inner_top   = hf*(1-np.cos(np.linspace(0,1,Nzf)*np.pi/2))
    zlocs         = np.concatenate([z_outer_bot, z_inner_bot, z_inner_top, z_outer_top])

    # generate spanwise grid point locations: placed between vortex lines to avoid discontinuities
    ypts        = VD.YC[breaks]
    y_semispan  = ypts[0:int(len(ypts)/2)]
    
    if L>=1.:
        # add grid points outside wingtip
        y_outerspan = np.linspace(1.01,L,Nyo_half)*half_span
        y_semispan  = np.append(y_semispan, y_outerspan)
    else:
        # trim spanwise points to region of interest
        y_in       = y_semispan<(L*half_span)
        y_semispan = y_semispan[y_in]
        
    ylocs       = np.concatenate([np.flipud(-y_semispan),y_semispan])
    
    # declare new control points
    cp_YC = np.repeat(ylocs,len(zlocs)) 
    cp_ZC = np.tile(zlocs,len(ylocs))
    cp_XC = np.ones_like(cp_YC)*x_plane 
    
    grid_points = Data()
    grid_points.XC = cp_XC
    grid_points.YC = cp_YC
    grid_points.ZC = cp_ZC
    grid_points.yline = ylocs
    grid_points.zline = zlocs 
        
    return grid_points


## @ingroup Analyses-Atmospheric
# Atmospheric.py
#
# Created:  Feb 2015, T. MacDonald
# Modified: Feb 2016, A. Wendorff


# ----------------------------------------------------------------------
#  Imports
# ----------------------------------------------------------------------

from RCAIDE.Library.Attributes.Atmospheres.Atmosphere import Atmosphere
from RCAIDE.Framework.Analyses import Analysis


# ----------------------------------------------------------------------
#  Analysis
# ----------------------------------------------------------------------

## @ingroup Analyses-Atmospheric
class Atmospheric(Analysis):
    """This is the base class for atmospheric analyses. It contains functions
    that are built into the default class.
    
    Assumptions:
    None
    
    Source:
    N/A
    """
    def __defaults__(self):
        """This sets the default values for the analysis to function.
        
        Assumptions:
        None
        
        Source:
        N/A
        
        Inputs:
        None
        
        Outputs:
        None
        
        Properties Used:
        N/A
        """          
        atmo_data = Atmosphere()
        self.update(atmo_data)
        
        
    def compute_values(self,altitude):
        """This function is not implemented for the base class."""
        raise NotImplementedError



# RCAIDE/Methods/Geometry/Two_Dimensional/Cross_Section/Airfoil/__init__.py
# 

""" RCAIDE Package Setup
"""

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

from .compute_naca_4series        import compute_naca_4series 
from .compute_airfoil_properties  import compute_airfoil_properties
from .import_airfoil_dat          import import_airfoil_dat
from .import_airfoil_geometry     import import_airfoil_geometry 
from .import_airfoil_polars       import import_airfoil_polars
from .convert_airfoil_to_meshgrid import convert_airfoil_to_meshgrid


# RCAIDE/Library/Missions/Segments/converge_root.py
# 
# 
# Created:  Jul 2023, M. Clarke  

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
# Package imports 
import scipy.optimize
import numpy as np
import  sys

# ----------------------------------------------------------------------------------------------------------------------
# converge root
# ---------------------------------------------------------------------------------------------------------------------- 
def converge_root(segment):
    """Interfaces the mission to a numerical solver. The solver may be changed by using root_finder.

    Assumptions:
    N/A

    Source:
    N/A

    Inputs:
    segment                            [Data]
    segment.settings.root_finder       [Data]
    state.numerics.tolerance_solution  [Unitless]

    Outputs:
    state.unknowns                     [Any]
    segment.state.numerics.converged   [Unitless]

    Properties Used:
    N/A
    """       
    
    unknowns = segment.state.unknowns.pack_array()
    
    try:
        root_finder = segment.settings.root_finder
    except AttributeError:
        root_finder = scipy.optimize.fsolve 
    
    unknowns,infodict,ier,msg = root_finder( iterate,
                                         unknowns,
                                         args = segment,
                                         xtol = segment.state.numerics.tolerance_solution,
                                         maxfev = segment.state.numerics.max_evaluations,
                                         epsfcn = segment.state.numerics.step_size,
                                         full_output = 1)
    
    if ier!=1:
        print("Segment did not converge. Segment Tag: " + segment.tag)
        print("Error Message:\n" + msg)
        segment.state.numerics.converged = False
        segment.converged = False
    else:
        segment.state.numerics.converged = True
        segment.converged = True
                            
    return
    
# ---------------------------------------------------------------------------------------------------------------------- 
#  Helper Functions
# ---------------------------------------------------------------------------------------------------------------------- 
def iterate(unknowns, segment):
    
    """Runs one iteration of of all analyses for the mission.

    Assumptions:
    N/A

    Source:
    N/A

    Inputs:
    state.unknowns                [Data]
    segment.process.iterate       [Data]

    Outputs:
    residuals                     [Unitless]

    Properties Used:
    N/A
    """       
    if isinstance(unknowns,np.ndarray):
        segment.state.unknowns.unpack_array(unknowns)
    else:
        segment.state.unknowns = unknowns
        
    segment.process.iterate(segment)
    
    residuals = segment.state.residuals.pack_array()
        
    return residuals 


# RCAIDE/Library/Components/Propulsors/Converters/Compression_Nozzle.py
# (c) Copyright 2023 Aerospace Research Community LLC
# 
# Created:  Feb 2024, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 
 # RCAIDE imports   
from RCAIDE.Library.Components                      import Component  
from RCAIDE.Library.Methods.Propulsors.Converters.Compression_Nozzle.append_compression_nozzle_conditions import append_compression_nozzle_conditions

# ---------------------------------------------------------------------------------------------------------------------- 
#  Compression Nozzle 
# ---------------------------------------------------------------------------------------------------------------------- 
class Compression_Nozzle(Component):
    """
    A compression nozzle component model for propulsion systems that converts flow energy into pressure.

    Attributes
    ----------
    tag : str
        Identifier for the nozzle. Default is 'Nozzle'.
        
    polytropic_efficiency : float
        Efficiency of the compression process. Default is 1.0.
        
    pressure_ratio : float
        Ratio of outlet to inlet pressure. Default is 1.0.
        
    pressure_recovery : float
        Pressure recovery factor. Default is 1.0.
        
    compressibility_effects : bool
        Flag for including compressibility effects. Default is False.
        
    compression_levels : float
        Number of compression stages. Default is 0.0.
        
    theta : float
        Nozzle angle [rad]. Default is 0.0.

    Notes
    -----
    The Compression_Nozzle class models a nozzle that increases the pressure of the flow.
    The model assumes:
    * Pressure ratio and efficiency do not change with varying conditions
    * Flow is either subsonic or choked

    **Definitions**

    'Polytropic Efficiency'
        Measure of compression process efficiency accounting for real gas effects

    'Pressure Recovery'
        Ratio of actual to ideal pressure rise
        
    'Choked Flow'
        Condition where flow reaches sonic velocity at the throat

    References
    ----------
    [1] Mattingly, J. D., & Boyer, K. M. (2016). Elements of propulsion: Gas 
        turbines and rockets, second edition Jack D. Mattingly, Keith M. Boyer. 
        American Institute of Aeronautics and Astronautics.

    See Also
    --------
    RCAIDE.Library.Methods.Propulsors.Converters.Compression_Nozzle
    """

    def __defaults__(self): 
        """This sets the default values for the component to function.

        Assumptions:
            1. Pressure ratio and efficiency do not change with varying conditions.
            2. Subsonic or choked output.

        Source:
            Mattingly, J. D., & Boyer, K. M. (2016). Elements of propulsion: Gas turbines and rockets, second edition Jack D. Mattingly, Keith M. Boyer. American Institute of Aeronautics and Astronautics. 
        """    
        self.tag                             = 'Nozzle'
        self.polytropic_efficiency           = 1.0
        self.pressure_ratio                  = 1.0
        self.pressure_recovery               = 1.0
        self.compressibility_effects         = False 
        self.compression_levels              = 0.0
        self.theta                           = 0.0

    def append_operating_conditions(self,segment,propulsor): 
        propulsor_conditions =  segment.state.conditions.energy[propulsor.tag]
        append_compression_nozzle_conditions(self,segment,propulsor_conditions)
        return


# RCAIDE/Methods/Energy/Sources/Battery/Common/size_module_from_mass.py
# 
# 
# Created:  Jul 2023, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  METHOD
# ---------------------------------------------------------------------------------------------------------------------- -Common
def size_module_from_mass(battery_module):
    """
    Calculate the max energy and power based of the mass
    Assumptions:
    A constant value of specific energy and power

    Inputs:
    mass              [kilograms]
    battery_module.
      specific_energy [J/kg]               
      specific_power  [W/kg]

    Outputs:
     battery_module.
       maximum_energy
       maximum_power
       mass_properties.
        mass


    """     
    mass = battery_module.mass_properties.mass/battery_module.BMS_additional_weight_factor
    
    if battery_module.cell.mass == None: 
        n_series   = 1
        n_parallel = 1 
    else:
        n_cells    = int(mass/battery_module.cell.mass)
        n_series   = int(battery_module.maximum_voltage/battery_module.cell.maximum_voltage)
        n_parallel = int(n_cells/n_series)
        
    battery_module.maximum_energy                    = mass*battery_module.cell.specific_energy  
    battery_module.maximum_power                     = mass*battery_module.cell.specific_power
    battery_module.initial_maximum_energy            = battery_module.maximum_energy    
    battery_module.electrical_configuration.series   = n_series
    battery_module.electrical_configuration.parallel = n_parallel     



# RCAIDE/Library/Methods/Propulsors/Electric_Rotor_Propulsor/pack_electric_ducted_fan_residuals.py
# (c) Copyright 2023 Aerospace Research Community LLC
# 
# Created:  Jun 2024, M. Clarke   

# ---------------------------------------------------------------------------------------------------------------------- 
#  pack electric ducted_fan network residuals
# ----------------------------------------------------------------------------------------------------------------------  

def pack_electric_ducted_fan_residuals(propulsor,segment):  
    motor         = propulsor.motor
    ducted_fan    = propulsor.ducted_fan 
    q_motor       = segment.state.conditions.energy[propulsor.tag][motor.tag].torque
    q_ducted_fan  = segment.state.conditions.energy[propulsor.tag][ducted_fan.tag].torque 
    segment.state.residuals.network[propulsor.tag  + '_ducted_fan_motor_torque'] = q_motor - q_ducted_fan
    return 



# RCAIDE/Library/Methods/Weights/Correlation_Buildups/BWB/compute_aft_centerbody_weight.py
# 
# Created: Sep 2024, M. Clarke 

# ---------------------------------------------------------------------------------------------------------------------- 
#  Imports
# ----------------------------------------------------------------------------------------------------------------------
from RCAIDE.Framework.Core import Units

# ---------------------------------------------------------------------------------------------------------------------- 
# Aft Centerbody Weight 
# ----------------------------------------------------------------------------------------------------------------------
def compute_aft_centerbody_weight(no_of_engines, aft_centerbody_area, aft_centerbody_taper, TOGW):
    """ Weight estimate for the aft section of a BWB centerbody.
        Regression from FEA by K. Bradley (George Washington University).
        
        Assumptions:
            -The engines are mounted on the aft centerbody
            -The aft centerbody is unpressurized
        
        Sources:
            Bradley, K. R., "A Sizing Methodology for the Conceptual Design of 
            Blended-Wing-Body Transports," NASA/CR-2004-213016, 2004.
            
        Inputs:
            no_of_engines - the number of engines mounted on the aft centerbody 
            [dimensionless]
            aft_centerbody_area - the planform area of the aft centerbody. 
            Typcially the area behind 70% chord [meters**2]
            aft_centerbody_taper - the taper ratio of the aft centerbody (exclude
            the chord taken up by the pressurized passenger cabin) [dimensionless]
            TOGW - Takeoff gross weight of the aircraft [kilograms]
        Outputs:
            W_aft - the estimated structural weight of the BWB aft centerbody
                
        Properties Used:
        N/A
        """         
    # convert to imperial units and shorten variable names 
    S_aft  = aft_centerbody_area  / Units.feet ** 2.0
    l_aft  = aft_centerbody_taper
    W      = TOGW/ Units.pounds
    
    W_aft = (1.0 + 0.05*no_of_engines) * 0.53 * S_aft * (W**0.2) * (l_aft + 0.5)
    
    # convert back to base units
    W_aft = W_aft * Units.pounds
    
    return W_aft


# 

""" RCAIDE Package Setup
"""

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 

from .append_battery_conditions               import append_battery_conditions, append_battery_segment_conditions
from .find_ragone_properties                  import find_ragone_properties
from .find_specific_power                     import find_specific_power
from .find_mass_gain_rate                     import find_mass_gain_rate
from .find_total_mass_gain                    import find_total_mass_gain
from .size_module_from_mass                    import size_module_from_mass
from .size_module_from_energy_and_power        import size_module_from_energy_and_power
from .compute_module_properties               import compute_module_properties 


# RCAIDE/Library/Methods/Aerodynamics/Athena_Vortex_Lattice/write_input_deck.py
#
# Created: Oct 2024, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

# RCAIDE imports 
from RCAIDE.Library.Components.Wings.Control_Surfaces import Aileron , Elevator , Rudder  
from RCAIDE.Framework.Core import Units
from .purge_files import purge_files

# ----------------------------------------------------------------------------------------------------------------------
#  write_input_deck
# ---------------------------------------------------------------------------------------------------------------------- 
def write_input_deck(avl_object,trim_aircraft,control_surfaces):
    """ This function writes the execution steps used in the AVL call
    Assumptions:
        None
        
    Source:
        Drela, M. and Youngren, H., AVL, http://web.mit.edu/drela/Public/web/avl
    Inputs:
        avl_object
    Outputs:
        None
 
    Properties Used:
        N/A
    """     
    mass_file_input = \
'''MASS {0}
mset
0
PLOP
G

'''   
    open_runs = \
'''CASE {}
'''
    base_input = \
'''OPER
'''
    # unpack
    batch         = avl_object.current_status.batch_file
    deck_filename = avl_object.current_status.deck_file 
    mass_filename = avl_object.settings.filenames.mass_file

    # purge old versions and write the new input deck
    purge_files([deck_filename]) 
    with open(deck_filename,'w') as input_deck:
        input_deck.write(mass_file_input.format(mass_filename))
        input_deck.write(open_runs.format(batch))
        input_deck.write(base_input)
        for case in avl_object.current_status.cases:
            # write and store aerodynamic and static stability result files 
            case_command = make_case_command(avl_object,case,trim_aircraft,control_surfaces)
            input_deck.write(case_command)

        input_deck.write('\nQUIT\n')

    return


def make_case_command(avl_object,case,trim_aircraft,control_surfaces):
    """ Makes commands for case execution in AVL
    Assumptions:
        None
        
    Source:
        None
    Inputs:
        case.index
        case.tag
        case.result_filename
    Outputs:
        case_command
 
    Properties Used:
        N/A
    """  
    # This is a template (place holder) for the input deck. Think of it as the actually keys
    # you will type if you were to manually run an analysis
    base_case_command = \
'''{0}{1}{2}{3}{4}
x 
{5}
{6}
{7}
{8}
{9}
{10}
{11}
{12}
'''  
    
    # if trim analysis is specified, this function writes the trim commands else it 
    # uses the defined deflection of the control surfaces of the aircraft
    if trim_aircraft:
        trim_command       = make_trim_text_command(case, avl_object)
        beta_command       = make_beta_text_command(case)
        roll_rate_command  = make_roll_rate_text_command(case)
        pitch_rate_command = make_pitch_rate_text_command(case)
    else:
        roll_rate_command  = ''
        pitch_rate_command = ''
        beta_command       = ''
        if control_surfaces:
            trim_command = control_surface_deflection_command(case,avl_object)
        else: 
            trim_command = ''
    
    index          = case.index
    case_tag       = case.tag
    
    # AVL executable commands which correlate to particular result types 
    aero_command_1 = 'st' # stability axis derivatives   
    aero_command_2 = 'fn' # surface forces 
    aero_command_3 = 'fs' # strip forces 
    aero_command_4 = 'sb' # body axis derivatives 
                   
    # create aliases for filenames for future handling
    aero_file_1    = case.aero_result_filename_1 
    aero_file_2    = case.aero_result_filename_2 
    aero_file_3    = case.aero_result_filename_3 
    aero_file_4    = case.aero_result_filename_4
    
    # purge files 
    if not avl_object.settings.keep_files:
        purge_files([aero_file_1])
        purge_files([aero_file_2])
        purge_files([aero_file_3])      
    
    # write input deck for avl executable 
    case_command = base_case_command.format(index,trim_command,roll_rate_command,pitch_rate_command ,beta_command,aero_command_1 , aero_file_1 ,aero_command_2  \
                                            , aero_file_2 , aero_command_3 , aero_file_3, aero_command_4 , aero_file_4) 
        
    return case_command

def make_trim_text_command(case,avl_object):
    """ Writes the trim command currently for a specified AoA or flight CL condition
    Assumptions:
        None
        
    Source:
        None
    Inputs:
        case
    Outputs:
        trim_command
 
    Properties Used:
        N/A
    """      
     
    cs_template = \
'''
D{0}
{1}
'''
    cs_idx = 1 
    cs_commands = ''
    for wing in avl_object.vehicle.wings:
        for ctrl_surf in wing.control_surfaces:
            if type(ctrl_surf) == Aileron:
                control = 'RM'
                cs_command = cs_template.format(cs_idx,control)
                cs_commands = cs_commands + cs_command
                cs_idx += 1   
            elif type(ctrl_surf) == Elevator:
                control = 'PM' 
                cs_command = cs_template.format(cs_idx,control)
                cs_commands = cs_commands + cs_command
                cs_idx += 1   
            elif type(ctrl_surf) == Rudder:
                control = 'YM' 
                cs_command = cs_template.format(cs_idx,control)
                cs_commands = cs_commands + cs_command
                cs_idx += 1   
            else:
                cs_idx += 1   
    base_trim_command = \
'''
c1
{0}
{1}
'''     
    # if Angle of Attack command is specified, write A 
    if case.conditions.aerodynamics.coefficients.lift.total == None:
        condition = 'A'
        val       = case.conditions.aerodynamics.angles.alpha
    elif case.conditions.aerodynamics.coefficients.lift.total > 0:  
        condition = 'C'
        val       = case.conditions.aerodynamics.coefficients.lift.total 
    else:   
        trim_command = ''
        return trim_command
        
    # write trim commands into template 
    trim_command = base_trim_command.format(condition,val)
    
    return cs_commands + trim_command

def make_roll_rate_text_command(case):
    """ Writes the roll rate command currently for a specified flight  condition
    Assumptions:
        None
        
    Source:
        None
    Inputs:
        case
    Outputs:
        trim_command
 
    Properties Used:
        N/A
    """       
    base_roll_command = \
'''
R
R
{0}'''      
    roll_rate_coeff  = case.conditions.static_stability.coefficients.roll  
    if  roll_rate_coeff != 0.0:
        roll_command     = base_roll_command.format(roll_rate_coeff)
    else:
        roll_command = ''
    return roll_command

def make_pitch_rate_text_command(case):
    """ Writes the pitch rate command currently for a specified flight  condition
    Assumptions:
        None
        
    Source:
        None
    Inputs:
        case
    Outputs:
        trim_command
 
    Properties Used:
        N/A
    """       
    base_pitch_command = \
'''
Y
Y
{0}'''       
    pitch_rate_coeff   = case.conditions.static_stability.coefficients.pitch  
    if pitch_rate_coeff != 0.0:
        pitch_command = base_pitch_command.format(pitch_rate_coeff)
    else:
        pitch_command = ''
    return pitch_command

def make_beta_text_command(case):
    """ Writes the roll rate command currently for a specified flight  condition
    Assumptions:
        None
        
    Source:
        None
    Inputs:
        case
    Outputs:
        trim_command
 
    Properties Used:
        N/A
    """       
    base_roll_command = \
'''
B
B
{0}'''      
    beta = case.conditions.aerodynamics.angles.beta 
    if  beta != 0.0:
        beta_command     = base_roll_command.format(beta)
    else:
        beta_command = ''
    return beta_command

def control_surface_deflection_command(case,avl_object): 
    """Writes the control surface command template
    Assumptions:
        None
        
    Source:
        None
    Inputs:
        avl_object
        case
    Outputs:
        em_case_command
 
    Properties Used:
        N/A
    """     
    cs_template = \
'''
D{0}
D{1}
{2}'''
    cs_idx = 1 
    cs_commands = ''
    for wing in avl_object.vehicle.wings:
        for ctrl_surf in wing.control_surfaces:
            cs_command = cs_template.format(cs_idx,cs_idx,round(ctrl_surf.deflection/Units.degrees,4))
            cs_commands = cs_commands + cs_command
            cs_idx += 1
    return cs_commands 


# RCAIDE/Library/Components/Propulsors/Converters/Propeller.py
# 
# 
# Created:  Mar 2024, M. Clarke 
# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------   
 # RCAIDE imports 
from .Rotor import Rotor

# ----------------------------------------------------------------------------------------------------------------------
#  Propeller
# ----------------------------------------------------------------------------------------------------------------------     
class Propeller(Rotor):
    """
    A propeller component model for aircraft propulsion, inheriting from the base Rotor class.

    Attributes
    ----------
    tag : str
        Identifier for the propeller. Default is 'propeller'.
        
    orientation_euler_angles : list
        Vector of angles [rad] defining propeller orientation [θ, φ, ψ]. 
        Default is [0., 0., 0.] for X-direction thrust in vehicle frame.
        
    use_2d_analysis : bool
        Flag for using 2D aerodynamic analysis. Default is False.
        
    variable_pitch : bool
        Flag indicating if propeller has variable pitch capability. Default is False.

    Notes
    -----
    The Propeller class models fixed or variable pitch propellers for aircraft
    propulsion. It includes capabilities for:

    * Forward flight performance analysis
    * Fixed and variable pitch operation
    * Blade element momentum theory calculations
    * Efficiency optimization
    * Acoustic analysis

    The model assumes:

    * Axial inflow (unless modified by orientation angles)
    * Rigid blades
    * Quasi-steady aerodynamics
    * No blade-to-blade aerodynamic interference
    * No compressibility effects (unless explicitly modeled)

    **Definitions**
    
    'Variable Pitch'
        Capability to change blade pitch angle during operation
    'Euler Angles'
        Set of three angles defining propeller orientation relative to vehicle frame
    '2D Analysis'
        Use of two-dimensional airfoil data for blade section analysis

    See Also
    --------
    RCAIDE.Library.Components.Propulsors.Converters.Rotor
    RCAIDE.Library.Components.Propulsors.Converters.Prop_Rotor
    """     
    def __defaults__(self):
        """This sets the default values for the component to function.
        
        Assumptions:
        None
        
        Source:
        N/A
        
        Inputs:
        None
        
        Outputs:
        None
        
        Properties Used:
        None
        """         

        self.tag                       = 'propeller'
        self.orientation_euler_angles  = [0.,0.,0.] # This is X-direction thrust in vehicle frame
        self.use_2d_analysis           = False       
        self.variable_pitch            = False
        



# RCAIDE/Library/Methods/Aerodynamics/Vortex_Lattice_Method/evaluate_AVL_surrogate.py
#  
# Created: Oct 2024, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

# RCAIDE imports   
from RCAIDE.Framework.Core import  Data   
from RCAIDE.Library.Methods.Aerodynamics.Athena_Vortex_Lattice.run_AVL_analysis  import run_AVL_analysis  

# package imports
import numpy   as np  

# ----------------------------------------------------------------------------------------------------------------------
#  Vortex_Lattice
# ---------------------------------------------------------------------------------------------------------------------- 
def evaluate_AVL_surrogate(state,settings,vehicle):
    """Evaluates surrogates forces and moments using built surrogates 
    
    Assumptions:
        None
        
    Source:
        None

    Args:
        aerodynamics : VLM analysis  [unitless]
        state        : flight conditions     [unitless]
        settings     : VLM analysis settings [unitless]
        vehicle      : vehicle configuration [unitless] 
        
    Returns: 
        None  
    """          
    conditions          = state.conditions
    aerodynamics        = state.analyses.aerodynamics  
    Mach                = conditions.freestream.mach_number
    AoA                 = conditions.aerodynamics.angles.alpha
    lift_model          = aerodynamics.surrogates.lift_coefficient            
    drag_model          = aerodynamics.surrogates.drag_coefficient            
    e_model             = aerodynamics.surrogates.span_efficiency_factor      
    moment_model        = aerodynamics.surrogates.moment_coefficient          
    Cm_alpha_model      = aerodynamics.surrogates.Cm_alpha_moment_coefficient 
    Cn_beta_model       = aerodynamics.surrogates.Cn_beta_moment_coefficient       
    neutral_point_model = aerodynamics.surrogates.neutral_point               
    cg                  = vehicle.mass_properties.center_of_gravity[0]
    MAC                 = vehicle.wings.main_wing.chords.mean_aerodynamic
  
    pts   = np.hstack((AoA,Mach))     
    conditions.aerodynamics.coefficients.lift.total                   = np.atleast_2d(lift_model(pts)).T  
    conditions.aerodynamics.coefficients.drag.induced.inviscid        = np.atleast_2d(drag_model(pts)).T  
    conditions.aerodynamics.span_efficiency                           = np.atleast_2d(e_model(pts)).T  
    conditions.control_surfaces.slat.static_stability.coefficients.M  = np.atleast_2d(moment_model(pts)).T  
    conditions.static_stability.derivatives.CM_alpha                  = np.atleast_2d(Cm_alpha_model(pts)).T  
    conditions.static_stability.derivatives.CN_beta                   = np.atleast_2d(Cn_beta_model(pts)).T  
    conditions.static_stability.neutral_point                         = np.atleast_2d(neutral_point_model(pts)).T    
    conditions.static_stability.static_margin                         = (conditions.static_stability.neutral_point - cg)/MAC     
    aerodynamics.settings.span_efficiency                             = conditions.aerodynamics.span_efficiency   
    return


def evaluate_AVL_no_surrogate(state,settings,vehicle):
    """Evaluates forces and moments directly using VLM.
    
    Assumptions:
        None
        
    Source:
        None

    Args:
        aerodynamics    : AVL analysis  [unitless]
        state           : flight conditions     [unitless] 
        vehicle         : vehicle configuration [unitless] 
        
    Returns: 
        None  
    """          

    # unpack 
    conditions     = state.conditions
    aerodynamics   = state.analyses.aerodynamics   
    run_AVL_analysis(aerodynamics,conditions)
                       
    return

 
 


# RCAIDE/Energy/Peripherals/Payload.py
# 
# 
# Created:  Jul 2023, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------   
# RCAIDE imports  
from RCAIDE.Library.Components import Component  
from RCAIDE.Library.Methods.Propulsors.Common.append_payload_conditions import append_payload_conditions
 
# ----------------------------------------------------------------------------------------------------------------------
#  Avionics
# ----------------------------------------------------------------------------------------------------------------------              
class Payload(Component):
    """
    A payload component model for representing mission cargo and equipment.

    Attributes
    ----------
    tag : str
        Identifier for the payload. Default is 'payload'.
        
    power_draw : float
        Power consumption of the payload. Default is 0.0.

    Notes
    -----
    The Payload class models mission-specific cargo and equipment, including:

    * Mass properties
    * Power requirements
    * Location in vehicle
    * Energy consumption
    * Operating conditions

    The model can represent:

    * Scientific instruments
    * Cargo containers
    * Passenger accommodations
    * Mission-specific equipment
    * Sensor packages
    * Communication systems

    **Major Assumptions**
    
    * Constant power draw during operation
    * Fixed mass and volume
    * Rigid mounting to vehicle
    * No thermal effects on vehicle
    * No aerodynamic effects
    * Steady-state operation

    See Also
    --------
    RCAIDE.Library.Components.Component
    RCAIDE.Library.Methods.Propulsors.Common.append_payload_conditions
    """          
    def __defaults__(self):
        """This sets the default power draw.

        Assumptions:
        None

        Source:
        N/A

        Inputs:
        None

        Outputs:
        None

        Properties Used:
        N/A
        """             
        self.tag        = 'payload' 
        self.power_draw = 0.0
         
    def append_operating_conditions(self,segment,bus): 
        append_payload_conditions(self,segment,bus)
        return 
        
    def power(self):
        """This gives the power draw from a payload.

        Assumptions:
        None

        Source:
        N/A

        Inputs:
        None

        Outputs:
        self.outputs.power_draw         [Watts]

        Properties Used:
        self.power_draw
        """          
        self.inputs.power = self.power_draw
        
        return self.power_draw 


# RCAIDE/Compoments/Fuselages/Fuselage.py
# 
# Created:  Mar 2024, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 
# RCAIDE imports
import RCAIDE
from RCAIDE.Framework.Core                import Data 
from RCAIDE.Library.Components.Component  import Container
from RCAIDE.Library.Components            import Component
from RCAIDE.Library.Methods.Weights.Moment_of_Inertia.compute_fuselage_moment_of_inertia import  compute_fuselage_moment_of_inertia

# ---------------------------------------------------------------------------------------------------------------------- 
#  Fuselage
# ---------------------------------------------------------------------------------------------------------------------- 
class Fuselage(Component):
    """
    Base class for aircraft fuselage components providing core functionality for 
    geometric definition and analysis.

    Attributes
    ----------
    tag : str
        Unique identifier for the fuselage component, defaults to 'fuselage'
        
    origin : list
        3D coordinates [x, y, z] defining the fuselage's reference point, 
        defaults to [[0.0, 0.0, 0.0]]
        
    aerodynamic_center : list
        3D coordinates [x, y, z] of the fuselage's aerodynamic center, 
        defaults to [0.0, 0.0, 0.0]
        
    differential_pressure : float
        Cabin pressurization differential, defaults to 0.0
        
    seats_abreast : float
        Number of passenger seats arranged side-by-side, defaults to 0.0
        
    seat_pitch : float
        Longitudinal distance between seat rows in meters, defaults to 0.0
        
    number_coach_seats : float
        Total number of economy class seats, defaults to 0.0
        
    areas : Data
        Collection of fuselage area measurements
        
        - front_projected : float
            Frontal area, defaults to 0.0
        - side_projected : float
            Side profile area, defaults to 0.0
        - wetted : float
            Total wetted surface area, defaults to 0.0
            
    effective_diameter : float
        Equivalent circular diameter, defaults to 0.0
        
    width : float
        Maximum fuselage width, defaults to 0.0
        
    heights : Data
        Collection of height measurements along the fuselage
        
        - maximum : float
            Maximum height in meters, defaults to 0.0
        - at_quarter_length : float
            Height at 25% of fuselage length, defaults to 0.0
        - at_three_quarters_length : float
            Height at 75% of fuselage length, defaults to 0.0
        - at_wing_root_quarter_chord : float
            Height at wing root quarter chord, defaults to 0.0
        - at_vertical_root_quarter_chord : float
            Height at vertical tail root quarter chord, defaults to 0.0
            
    lengths : Data
        Collection of length measurements
        
        - nose : float
            Length of nose section, defaults to 0.0
        - tail : float
            Length of tail section, defaults to 0.0
        - total : float
            Total fuselage length, defaults to 0.0
        - cabin : float
            Length of passenger cabin, defaults to 0.0
        - fore_space : float
            Length of forward equipment bay, defaults to 0.0
        - aft_space : float
            Length of aft equipment bay, defaults to 0.0
            
    x_rotation : float
        Rotation angle about x-axis, defaults to 0.0
        
    y_rotation : float
        Rotation angle about y-axis, defaults to 0.0
        
    z_rotation : float
        Rotation angle about z-axis, defaults to 0.0
        
    fineness : Data
        Fineness ratio parameters
        
        - nose : float
            Ratio of nose length to maximum diameter, defaults to 0.0
        - tail : float
            Ratio of tail length to maximum diameter, defaults to 0.0
            
    nose_curvature : float
        Shape parameter defining nose profile curvature, defaults to 1.5
        
    tail_curvature : float
        Shape parameter defining tail profile curvature, defaults to 1.5
        
    fuel_tanks : Container
        Collection of fuel tank components within the fuselage, initialized empty
        
    vsp_data : Data
        OpenVSP-specific geometry parameters
        
        - xsec_surf_id : str
            OpenVSP cross-section surface identifier
        - xsec_num : int
            Number of cross-sections in OpenVSP model, defaults to None
            
    segments : Container
        Collection of fuselage segment components, initialized empty

    Notes
    -----
    This class serves as the foundation for all fuselage types in RCAIDE, providing:
    
    * Geometric definition capabilities
    * Integration with OpenVSP for visualization and analysis
    * Support for fuel tank and segment management
    * Basic structural configuration parameters
    
    **Definitions**

    'Fineness Ratio'
        Ratio of length to maximum diameter, important for drag characteristics
        
    'Wetted Area'
        Total surface area exposed to airflow
        
    'Equipment Bay'
        Non-pressurized volume for systems and equipment storage

    See Also
    --------
    RCAIDE.Library.Components.Fuselages.Tube_Fuselage
        Implementation for conventional tube-and-wing aircraft
    RCAIDE.Library.Components.Fuselages.Blended_Wing_Body_Fuselage
        Implementation for blended wing body aircraft
    """
    
    def __defaults__(self):
        """
        Sets default values for all fuselage attributes.
        """      
        
        self.tag                                    = 'fuselage'
        self.origin                                 = [[0.0,0.0,0.0]]
        self.aerodynamic_center                     = [0.0,0.0,0.0] 
        self.differential_pressure                  = 0.0    
        self.seats_abreast                          = 0.0
        self.seat_pitch                             = 0.0
        self.number_coach_seats                     = 0.0

        self.areas                                  = Data()
        self.areas.front_projected                  = 0.0
        self.areas.side_projected                   = 0.0
        self.areas.wetted                           = 0.0
        
        self.effective_diameter                     = 0.0
        self.width                                  = 0.0  
        
        self.heights                                = Data() 
        self.heights.maximum                        = 0.0
        self.heights.at_quarter_length              = 0.0
        self.heights.at_three_quarters_length       = 0.0
        self.heights.at_wing_root_quarter_chord     = 0.0
        self.heights.at_vertical_root_quarter_chord = 0.0 
        
        self.lengths                                = Data()     
        self.lengths.nose                           = 0.0
        self.lengths.tail                           = 0.0
        self.lengths.total                          = 0.0 
        self.lengths.cabin                          = 0.0 
        self.lengths.fore_space                     = 0.0
        self.lengths.aft_space                      = 0.0 
        
        self.x_rotation                             = 0.0
        self.y_rotation                             = 0.0
        self.z_rotation                             = 0.0 

        self.fineness                               = Data() 
        self.fineness.nose                          = 0.0 
        self.fineness.tail                          = 0.0  
        self.nose_curvature                         = 1.5
        self.tail_curvature                         = 1.5   
    
        self.fuel_tanks                             = Container()
 
        self.vsp_data                               = Data()
        self.vsp_data.xsec_surf_id                  = ''    # There is only one XSecSurf in each VSP geom.
        self.vsp_data.xsec_num                      = None  # Number if XSecs in fuselage geom. 
        self.segments                               = Container()

        self.vsp_data                               = Data()
        self.vsp_data.xsec_id                       = ''       
        self.vsp_data.shape                         = ''                
        
    def append_segment(self,segment):
        """
        Adds a new segment to the fuselage's segment container.

        Parameters
        ----------
        segment : Data
            Fuselage segment to be added
        """

        # Assert database type
        if not isinstance(segment,RCAIDE.Library.Components.Fuselages.Segments.Segment):
            raise Exception('input component must be of type Segment')

        # Store data
        self.segments.append(segment)

        return
    
    def append_fuel_tank(self,fuel_tank):
        """
        Adds a new fuel tank to the fuselage's fuel tank container.

        Parameters
        ----------
        fuel_tank : Data
            Fuel tank component to be added
        """

        # Assert database type
        if not isinstance(fuel_tank,Data):
            raise Exception('input component must be of type Data()')
    
        # Store data
        self.Fuel_Tanks.append(fuel_tank)

        return 

    def compute_moment_of_inertia(self, center_of_gravity=[[0, 0, 0]]): 
        """
        Computes the moment of inertia tensor for the fuselage.

        Parameters
        ----------
        center_of_gravity : list, optional
            Reference point coordinates for moment calculation, defaults to [[0, 0, 0]]

        Returns
        -------
        I : ndarray
            3x3 moment of inertia tensor in kg*m^2

        See Also
        --------
        RCAIDE.Library.Methods.Weights.Moment_of_Inertia.compute_fuselage_moment_of_inertia
            Implementation of the moment of inertia calculation
        """
        I = compute_fuselage_moment_of_inertia(self,center_of_gravity) 
        return I    


# RCAIDE/Library/Compoments/Mass_Properties.py
# 
# Created:  Mar 2024, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------   
# RCAIDE Imports 
from RCAIDE.Framework.Core import Data

# package imports 
import numpy as np

# ----------------------------------------------------------------------------------------------------------------------
#  Mass_Properties
# ----------------------------------------------------------------------------------------------------------------------        
class Mass_Properties(Data):
    """
    A class containing mass and inertial properties for physical components.

    Attributes
    ----------
    mass : float
        Total mass of the component, defaults to 0.0
        
    volume : float
        Volume of the component, defaults to 0.0
        
    center_of_gravity : ndarray
        3D coordinates [x, y, z] of the component's center of gravity, 
        defaults to [0.0, 0.0, 0.0]
        
    moments_of_inertia : Data
        Collection of inertial properties
        
        - center : ndarray
            Reference point for inertia calculations [x, y, z],
            defaults to [0.0, 0.0, 0.0]
        - tensor : ndarray
            3x3 inertia tensor about the reference point,
            defaults to zero matrix

    Notes
    -----
    The Mass_Properties class provides a standardized structure for tracking mass-related 
    properties of components. It includes:
    
    * Basic mass and volume
    * Center of gravity location
    * Full inertia tensor definition
    * Reference point specification

    See Also
    --------
    RCAIDE.Library.Components.Component
        Parent class that uses Mass_Properties
    RCAIDE.Framework.Core.Data
        Base class providing data structure functionality
    """
    def __defaults__(self):
        """
        Sets default values for mass property attributes.
        """         
        self.mass   = 0.0
        self.volume = 0.0
        self.center_of_gravity = np.array([[0.0,0.0,0.0]])
        
        self.moments_of_inertia = Data()
        self.moments_of_inertia.center = np.array([0.0,0.0,0.0])
        self.moments_of_inertia.tensor = np.array([[0.0,0.0,0.0],
                                                  [0.0,0.0,0.0],
                                                  [0.0,0.0,0.0]])


## @ingroup Methods-Geometry-Two_Dimensional-Cross_Section-Planform
# vertical_tail_planform_raymer.py
#
# Created:  ### ####, M. Vegh
# Modified: Jan 2016, E. Botero


# ----------------------------------------------------------------------
#  Imports
# ----------------------------------------------------------------------

from RCAIDE.Library.Methods.Geometry.Planform  import wing_planform

# ----------------------------------------------------------------------
#  Methods
# ----------------------------------------------------------------------
## @ingroup Methods-Geometry-Two_Dimensional-Cross_Section-Planform
def vertical_tail_planform_raymer(vertical_stabilizer, wing,  l_vt,c_vt):
    """Adjusts reference area before calling generic wing planform function to compute wing planform values.

    Assumptions:
    None

    Source:
    Raymer

    Inputs:
    vertical_stabilizer                    [RCAIDE data structure]
    wing                                   [RCAIDE data structure]  (should be the main wing)
    l_vt                                   [m] length from wing mean aerodynamic chord (MAC) to horizontal stabilizer MAC
    c_vt                                   [-] horizontal tail coefficient (Raymer specific) .02 = Sailplane, .04 = homebuilt, 
                                               .04 = GA single engine, .07 = GA twin engine, .04 = agricultural, 
                                               .08 = twin turboprop, .06 = flying boat, .06 = jet trainer, .07 = jet fighter
                                               .08 = military cargo/bomber, .09 = jet transport

    Outputs:
    vertical_stabilier.areas.reference     [m^2]
    Other changes to vertical_stabilizer (see wing_planform)

    Properties Used:
    N/A
    """           

    vertical_stabilizer.areas.reference = wing.spans.projected*c_vt*wing.areas.reference/l_vt
  
    wing_planform(vertical_stabilizer)
    
    return 0
    



# RCAIDE/Methods/Thermal_Management/Batteries/Heat_Acquisition_System/Direct_Air_Heat_Acquisition/direct_convection_model.py
# 
# 
# Created:  Jul 2023, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
import numpy as np

# ----------------------------------------------------------------------------------------------------------------------
#  Compute Net Convected Heat 
# ---------------------------------------------------------------------------------------------------------------------- 
def air_cooled_performance(HAS,battery,bus,coolant_line, Q_heat_gen,T_cell,state,delta_t,t_idx):
    '''Computes the net heat removed by direct air heat acquisition system.

    Assumptions:
    1) Battery pack cell heat transfer can be modelled as a cooling columns in a cross-flow
    2) Isothermal battery cell - the temperature at the center of the cell is the same at 
    the surface of the cell

    Source: 
    Heat Transfer Model:
    Pakowski, Zdzisław. "Fundamentals of Heat and Mass Transfer, Frank P Incropera,
    David P DeWitt, Theodore L Bergman, Adrienne S Lavine, J. Wiley & Sons, Hoboken
    NJ (2007), 997 pp." (2007): 1683-1684.,  Chapter 7 pg 437-446 

    Inputs:  
              h                         (heat transfer coefficient)  [W/(m^2*K)] 
              As_cell                   (battery cell surface area)  [meters^2]
              H_cell                    (battery cell height)        [meters]
              T_ambient                 (ambient temperature)        [Kelvin]
              T_current                 (pack temperature)           [Kelvin]
              T_cell                    (battery cell temperature)   [Kelvin] 
              heat_transfer_efficiency                               [unitless]
              
      
      Outputs:
             T_cell                     (Updated battery cell temperature) [Kelvin]
  
    Properties Used:
    None 
    '''
    
    
    # battery properties  
    As_cell                  = battery.cell.surface_area 
    D_cell                   = battery.cell.diameter                     
    H_cell                   = battery.cell.height              
    cell_mass                = battery.cell.mass    
    Cp                       = battery.cell.specific_heat_capacity    
    Nn                       = battery.geometrtic_configuration.normal_count            
    Np                       = battery.geometrtic_configuration.parallel_count    
    n_total_module           = Nn*Np  
    h                        = HAS.convective_heat_transfer_coefficient 
    heat_transfer_efficiency = HAS.heat_transfer_efficiency   
    T_ambient                = state.conditions.freestream.temperature[t_idx,:] 
    
    if n_total_module == 1: 
        # Using lumped model   
        Q_convec       = h*As_cell*(T_cell - T_ambient)
        Q_heat_gen_tot = Q_heat_gen 

    else:   
        K_coolant                    = state.conditions.freestream.thermal_conductivity[t_idx,:]
        nu_coolant                   = state.conditions.freestream.kinematic_viscosity[t_idx,:]
        Pr_coolant                   = state.conditions.freestream.prandtl_number[t_idx,:]
        rho_coolant                  = state.conditions.freestream.density[t_idx,:]    
        Cp_coolant                   = HAS.cooling_fluid.compute_cp(state.conditions.freestream.temperature[t_idx,:],state.conditions.freestream.pressure[t_idx,:] )
        V_coolant                    = HAS.cooling_fluid.flowspeed  
        
        # Chapter 7 pg 437-446 of Fundamentals of heat and mass transfer 
        S_T             = battery.geometrtic_configuration.normal_spacing          
        S_L             = battery.geometrtic_configuration.parallel_spacing

        S_D = np.sqrt(S_T**2+S_L**2)
        if 2*(S_D-D_cell) < (S_T-D_cell):
            V_max = V_coolant*(S_T/(2*(S_D-D_cell)))
        else:
            V_max = V_coolant*(S_T/(S_T-D_cell))

        T        = (T_ambient+T_cell)/2   
        Re_max   = V_max*D_cell/nu_coolant   
        if all(Re_max) > 10E2: 
            C = 0.35*((S_T/S_L)**0.2) 
            m = 0.6 
        else:
            C = 0.51
            m = 0.5  

        Pr_w_coolant          = HAS.cooling_fluid.compute_prandtl_number(T)            
        Nu                    = C*(Re_max**m)*(Pr_coolant**0.36)*((Pr_coolant/Pr_w_coolant)**0.25)           
        h                     = Nu*K_coolant/D_cell
        Tw_Ti                 = (T - T_ambient)
        Tw_To                 = Tw_Ti * np.exp((-np.pi*D_cell*n_total_module*h)/(rho_coolant*V_coolant*Nn*S_T*Cp_coolant))
        dT_lm                 = (Tw_Ti - Tw_To)/np.log(Tw_Ti/Tw_To)
        Q_convec              = heat_transfer_efficiency*h*np.pi*D_cell*H_cell*n_total_module*dT_lm 
        Q_convec[Tw_Ti == 0.] = 0.
        Q_heat_gen_tot        = Q_heat_gen*n_total_module  
    Q_net                     = Q_heat_gen_tot - Q_convec  
    dT_dt                     = Q_net/(cell_mass*n_total_module*Cp)
    T_current                 = T_cell + dT_dt*delta_t
    heat_transfer_efficiency  = (Tw_To - T_ambient) / (T_cell - T_ambient)
    
    state.conditions.energy[coolant_line.tag][HAS.tag].total_heat_removed[t_idx+1]               = Q_convec
    state.conditions.energy[coolant_line.tag][HAS.tag].effectiveness[t_idx+1]                    = heat_transfer_efficiency
       
    
    return  T_current


# RCAIDE/Library/Methods/Thermal_Management/Batteries/Liquid_Cooled_Wavy_Channel/__init__.py

""" RCAIDE Package Setup
"""

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 
 

from .design_wavy_channel                     import design_wavy_channel
from .wavy_channel_geometry_setup             import wavy_channel_geometry_setup
from .wavy_channel_sizing_setup               import wavy_channel_sizing_setup
from .wavy_channel_rating_model               import wavy_channel_rating_model
from .append_wavy_channel_conditions          import append_wavy_channel_conditions
from .append_wavy_channel_conditions          import append_wavy_channel_segment_conditions


# RCAIDE/Library/Methods/Weights/Correlation_Buildups/Transport/compute_fuselage_weight.py
# 
# 
# Created:  Sep 2024, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

# RCAIDE 
from RCAIDE.Framework.Core    import Units

# python imports 
 
# ----------------------------------------------------------------------------------------------------------------------
# fuselage Weight 
# ----------------------------------------------------------------------------------------------------------------------
def compute_fuselage_weight(vehicle, fuselage, W_wing, W_propulsion):
    """ Calculate the weight of a fuselage in the state tube and wing configuration
    
    Assumptions:
        fuselage in a standard wing and tube configuration         
    
    Source: 
        N/A 
        
    Inputs:
        fuselage.areas.wetted - fuselage wetted area                                                            [meters**2]
        fuselage.differential_pressure- Maximum fuselage pressure differential                                  [Pascal]
        fuselage.width - width of the fuselage                                                                  [meters]
        fuselage.heights.maximum - height of the fuselage                                                       [meters]
        fuselage.lengths.total - length of the fuselage                                                         [meters]
        vehicle.flight_envelope.limit_load - limit load factor at zero fuel weight of the aircraft                 [dimensionless]
        vehicle.mass_properties.max_zero_fuel - zero fuel weight of the aircraft                            [kilograms]
        W_wing - weight of the wing of the aircraft                           [kilograms]
        W_propulsion - weight of the entire propulsion system of the aircraft                              [kilograms]
        vehicle.wings.main_wing.chords.root - wing root chord                                               [meters]
        
    Outputs:
        weight - weight of the fuselage                                                                     [kilograms]
            
    Properties Used:
        N/A
    """
    # unpack inputs

    differential_pressure  = fuselage.differential_pressure / (Units.force_pound / Units.ft ** 2)  # Convert Pascals to lbs/ square ft
    width                  = fuselage.width / Units.ft  # Convert meters to ft
    height                 = fuselage.heights.maximum / Units.ft  # Convert meters to ft

    # setup
    length  = fuselage.lengths.total - vehicle.wings.main_wing.chords.root / 2.
    length  = length / Units.ft  # Convert meters to ft
    weight  = (vehicle.mass_properties.max_zero_fuel - W_wing - W_propulsion) / Units.lb  # Convert kg to lbs
    area    = fuselage.areas.wetted / Units.ft ** 2  # Convert square meters to square ft

    # process

    # Calculate fuselage indices
    I_p = 1.5 * 10 ** -3. * differential_pressure * width
    I_b = 1.91 * 10 ** -4. * vehicle.flight_envelope.positive_limit_load* weight * length / height ** 2.

    if I_p > I_b:
        I_f = I_p
    else:
        I_f = (I_p ** 2. + I_b ** 2.) / (2. * I_b)

    # Calculate weight of wing for traditional aircraft vertical tail without rudder
    fuselage_weight = ((1.051 + 0.102 * I_f) * area) * Units.lb  # Convert from lbs to kg

    return fuselage_weight



# RCAIDE/Methods/Energy/Sources/Battery/Ragone/find_specific_power.py
# 
# 
# Created:  Jul 2023, M. Clarke 
 
# ----------------------------------------------------------------------------------------------------------------------
#  METHOD
# ---------------------------------------------------------------------------------------------------------------------- -Ragone
def find_specific_power(battery, specific_energy):
    """determines specific specific power from a ragone curve correlation
    Assumptions:
    None
    
    Inputs:
    battery.
      specific_energy [J/kg]               
      ragone.
        constant_1    [W/kg]
        constant_2    [J/kg]
                
    Outputs:
    battery.
      specific_power  [W/kg]   
    
    
    
    """
    
    const_1                 = battery.cell.ragone.const_1
    const_2                 = battery.cell.ragone.const_2
    specific_power          = const_1*10.**(const_2*specific_energy)
    battery.specific_power  = specific_power
    battery.specific_energy = specific_energy


# RCAIDE/Library/Methods/Propulsors/Electric_Rotor_Propulsor/pack_electric_rotor_residuals.py
# (c) Copyright 2023 Aerospace Research Community LLC
# 
# Created:  Jun 2024, M. Clarke   

# ---------------------------------------------------------------------------------------------------------------------- 
#  pack electric rotor network residuals
# ----------------------------------------------------------------------------------------------------------------------  

def pack_electric_rotor_residuals(propulsor,segment): 
    propulsor_results   = segment.state.conditions.energy
    motor               = propulsor.motor
    rotor               = propulsor.rotor 
    q_motor             = propulsor_results[propulsor.tag][motor.tag].torque
    q_prop              = propulsor_results[propulsor.tag][rotor.tag].torque 
    segment.state.residuals.network[ propulsor.tag + '_rotor_motor_torque'] = q_motor - q_prop 
    return 



# RCAIDE/Library/Missions/Common/Update/altitude.py
# 
# 
# Created:  Jul 2023, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  Update Altitude
# ---------------------------------------------------------------------------------------------------------------------- 
def altitude(segment):
    """ Updates freestream altitude from inertial position
        
        Assumptions:
        N/A
        
        Inputs:
            segment.state.conditions:
                frames.inertial.position_vector [meters]
        Outputs:
            segment.state.conditions:
                freestream.altitude             [meters]
      
        Properties Used:
        N/A
                    
    """    
    altitude = -segment.state.conditions.frames.inertial.position_vector[:,2]
    segment.state.conditions.freestream.altitude[:,0] = altitude 


# RCAIDE/Library/Missions/Segments/expand_state.py
# 
# 
# Created:  Jul 2023, M. Clarke   

# ----------------------------------------------------------------------------------------------------------------------
# Expand State
# ----------------------------------------------------------------------------------------------------------------------  
def expand_state(segment):
    
    """Makes all vectors in the state the same size.

    Assumptions:
    N/A

    Source:
    N/A

    Inputs:
    state.numerics.number_of_control_points  [Unitless]

    Outputs:
    N/A

    Properties Used:
    N/A
    """       

    n_points = segment.state.numerics.number_of_control_points
    
    segment.state.expand_rows(n_points)
    
    return
    


# RCAIDE/Methods/Aerodynamics/Common/Drag/__init__.py
# 

""" RCAIDE Package Setup
"""

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 

from .parasite_drag_wing                        import parasite_drag_wing
from .parasite_drag_pylon                       import parasite_drag_pylon
from .parasite_total                            import parasite_total
from .induced_drag                              import induced_drag
from .wave_drag                                 import wave_drag 
from .lift_wave_drag                            import lift_wave_drag 
from .parasite_drag_fuselage                    import parasite_drag_fuselage
from .compressibility_drag                      import compressibility_drag
from .miscellaneous_drag                        import miscellaneous_drag  
from .parasite_drag_nacelle                     import parasite_drag_nacelle   
from .spoiler_drag                              import spoiler_drag 
from .wave_drag                                 import wave_drag 
from .total_drag                                import total_drag 
from .drag_divergence                           import drag_divergence
from .supersonic_wave_drag_volume_raymer        import supersonic_wave_drag_volume_raymer
from .supersonic_wave_drag_volume_sears_haack   import supersonic_wave_drag_volume_sears_haack
from .cooling_drag                              import cooling_drag
from .windmilling_drag                          import windmilling_drag
from .asymmetry_drag                            import asymmetry_drag
from .wave_drag_lift                            import wave_drag_lift
from .compressible_mixed_flat_plate             import compressible_mixed_flat_plate
from .estimate_2ndseg_lift_drag_ratio           import estimate_2ndseg_lift_drag_ratio
from .compressible_turbulent_flat_plate         import compressible_turbulent_flat_plate
from .compressible_mixed_flat_plate             import compressible_mixed_flat_plate


# RCAIDE/Library/Components/Energy/Sources/Battery_Modules/Lithium_Sulfur.py
# 
# 
# Created:  Mar 2024, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 

# package imports
from RCAIDE.Framework.Core import Units
from .Generic_Battery_Module import Generic_Battery_Module

# ----------------------------------------------------------------------------------------------------------------------
#  Lithium_Sulfur
# ----------------------------------------------------------------------------------------------------------------------  
class Lithium_Sulfur(Generic_Battery_Module):
    """
    Class for modeling lithium-sulfur battery characteristics and performance
    
    Attributes
    ----------
    cell.specific_energy : float
        Energy capacity per unit mass [J/kg] (default: 500 Wh/kg)
        
    cell.specific_power : float
        Maximum power output per unit mass [W/kg] (default: 1 kW/kg)
        
    cell.ragone.const_1 : float
        First constant in Ragone curve fit [kW/kg] (default: 245.848 kW/kg)
        
    cell.ragone.const_2 : float
        Second constant in Ragone curve fit [1/(Wh/kg)] (default: -0.00478 kg/Wh)
        
    cell.ragone.lower_bound : float
        Minimum specific energy on Ragone plot [Wh/kg] (default: 300 Wh/kg)
        
    cell.ragone.upper_bound : float
        Maximum specific energy on Ragone plot [Wh/kg] (default: 700 Wh/kg)

    Notes
    -----
    Lithium-sulfur batteries offer higher specific energy than conventional
    lithium-ion batteries but typically have lower power density. The Ragone
    plot parameters define the relationship between specific power and energy.

    See Also
    --------
    RCAIDE.Library.Components.Energy.Sources.Battery_Modules.Generic_Battery_Module
        Base battery module class
    """
    
    def __defaults__(self):
        """
        Sets default values for lithium-sulfur battery attributes
        """   
        self.cell.specific_energy    = 500     *Units.Wh/Units.kg
        self.cell.specific_power     = 1       *Units.kW/Units.kg
        self.cell.ragone.const_1     = 245.848 *Units.kW/Units.kg
        self.cell.ragone.const_2     = -.00478 /(Units.Wh/Units.kg)
        self.cell.ragone.lower_bound = 300     *Units.Wh/Units.kg
        self.cell.ragone.upper_bound = 700     *Units.Wh/Units.kg


# RCAIDE/Library/Methods/Aerodynamics/Common/Drag/pparasite_drag_pylon.py
# (c) Copyright 2023 Aerospace Research Community LLC
# 
# Created:  Jun 2024, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 
# RCAIDE imporst 
from RCAIDE.Framework.Core import Data 

# python imports 
import numpy as np

# ---------------------------------------------------------------------------------------------------------------------- 
#  Pylon Drag Fuselage
# ----------------------------------------------------------------------------------------------------------------------   
def parasite_drag_pylon(state,settings,geometry):
    """Computes the parasite drag due to pylons as a proportion of the network 

    Assumptions:
        None 

    Source:
        Stanford AA241 A/B Course Notes

    Args:
        state.conditions.aerodynamics.coefficients.drag.parasite[network.tag].
          form_factor                      (float): form_factor              [Unitless]
          compressibility_factor           (float): compressibility_factor   [Unitless]
          skin_friction            (numpy.ndarray): skin_friction            [Unitless]
          wetted_area              (numpy.ndarray): wetted_area              [m^2]
          parasite_drag            (numpy.ndarray): parasite_drag            [Unitless]
          reynolds_number          (numpy.ndarray): reynolds_number          [Unitless]
        geometry                            (dict): aircraft geometry        [-]

    Returns:
        None 
    """
    
    drag          = state.conditions.aerodynamics.coefficients.drag
    pylon_factor  = 0.2  

    # Estimating pylon drag
    for network in  geometry.networks: 
        for propulsor in network.propulsors:  
            if 'nacelle' in propulsor:
                nacelle              =  propulsor.nacelle
                if nacelle.has_pylon:
                    ref_area             = nacelle.diameter**2 / 4 * np.pi
                    pylon_parasite_drag  = pylon_factor *  drag.parasite[nacelle.tag].total* (ref_area/geometry.reference_area)
                    pylon_wetted_area    = pylon_factor *  drag.parasite[nacelle.tag].wetted_area  
                    pylon_cf             = drag.parasite[nacelle.tag].skin_friction 
                    pylon_compr_fact     = drag.parasite[nacelle.tag].compressibility_factor
                    pylon_rey_fact       = drag.parasite[nacelle.tag].reynolds_factor
                    pylon_FF             = drag.parasite[nacelle.tag].form_factor 
                    pylon_result         = Data(
                        wetted_area               = pylon_wetted_area   ,
                        reference_area            = geometry.reference_area   ,
                        total                     = pylon_parasite_drag ,
                        skin_friction             = pylon_cf  ,
                        compressibility_factor    = pylon_compr_fact   ,
                        reynolds_factor           = pylon_rey_fact   ,
                        form_factor               = pylon_FF   , )
                    drag.parasite[ nacelle.tag + '_pylon'] = pylon_result
                else:
                    pylon_result = Data(
                        wetted_area               = 0 ,
                        reference_area            = geometry.reference_area  ,
                        total                     = np.zeros_like(drag.parasite[nacelle.tag].skin_friction)  ,
                        skin_friction             = 0 ,
                        compressibility_factor    = 0 ,
                        reynolds_factor           = 0 ,
                        form_factor               = 0 , )
                    drag.parasite[ nacelle.tag + '_pylon'] = pylon_result
    return 


# RCAIDE/Framework/Analyses/Process.py
# 
# 
# Created:  Jul 2023, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------   
from RCAIDE.Framework.Core import ContainerOrdered
from RCAIDE.Framework.Core import Data 

# ----------------------------------------------------------------------------------------------------------------------
# Process
# ----------------------------------------------------------------------------------------------------------------------  
class Process(ContainerOrdered):
    """ RCAIDE.Framework.Analyses.Process()
    
        The Top Level Process Container Class
        
            Assumptions:
            None
            
            Source:
            N/A
    """     
    
    def evaluate(self,*args,**kwarg):
        """This is used to execute the evaluate functions of the analyses
            stored in the container.
        
                Assumptions:
                None
        
                Source:
                N/A
        
                Inputs:
                None
        
                Outputs:
                Results of the Evaluate Functions
        
                Properties Used:
                N/A
            """        
        
        results = Data() 
        
        for tag,step in self.items():  
            
            if hasattr(step,'evaluate'): 
                result = step.evaluate(*args,**kwarg)
            else:
                result = step(*args,**kwarg)
                
            results[tag] = result
         
        return results
        
    def __call__(self,*args,**kwarg):
        """This is used to set the class' call behavior to the evaluate functions.
        
                Assumptions:
                None
        
                Source:
                N/A
        
                Inputs:
                None
        
                Outputs:
                None
        
                Properties Used:
                N/A
            """                        
        return self.evaluate(*args,**kwarg) 
    



# lead-calculations.py
#
# Created:  Feb 2020, K. Hamilton - Through New Zealand Ministry of Business Innovation and Employment Research Contract RTVU2004 
# Modified: Nov 2021, S. Claridge

# ----------------------------------------------------------------------
#  Imports
# ----------------------------------------------------------------------

# RCAIDE imports 
from scipy import integrate 
from scipy.misc import derivative 

# ----------------------------------------------------------------------
#  Compute Minimum Power 
# ----------------------------------------------------------------------
def compute_minimum_power(material, cold_temp, hot_temp, current):
    """Calculate minimum electrical power

    Assumptions:
    N/A

    Source:
    N/A

    Inputs:
        material
        cold_temp    [K]
        hot_temp     [K]
        current      [A]

    Outputs:
        power        [W]    

    Properties Used:
    N/A
    """    

    # Estimate the area under the thermal:electrical conductivity vs temperature plot for the temperature range of the current lead.
    integral = integrate.quad(lambda T: material.thermal_conductivity(T)/material.electrical_conductivity(T), cold_temp, hot_temp)

    # Estimate the average thermal:electrical conductivity for the lead.
    average_ratio = (1/(hot_temp-cold_temp)) * integral[0]

    # Solve the heat flux at the cold end. This is both the load on the cryocooler and the power loss in the current lead.
    minimum_Q = current * (2*average_ratio*(hot_temp-cold_temp))**0.5

    # This represents the special case where all the electrical power is delivered to the cryogenic environment as this optimised the lead for reduced cryogenic load. Q = electrical power
    power = minimum_Q

    return power

# ----------------------------------------------------------------------
#  compute_optimal_ratio
# ----------------------------------------------------------------------
def compute_optimal_ratio( material, cold_temp, hot_temp, current, minimum_Q):
    """Calculate the optimum length to cross-sectional area ratio

    Assumptions:
    N/A

    Source:
    N/A

    Inputs:
        material
        cold_temp    [K]
        hot_temp     [K]
        current      [A]
        minimum_Q    [W]

    Outputs:
        la_ratio

    Properties Used:
    N/A
    """  
    # Calculate the optimum length to cross-sectional area ratio
    # Taken directly from McFee
    sigTL = material.electrical_conductivity(cold_temp)
    inte = integrate.quad(lambda T: compute_minimum_power(material,T,hot_temp,current)*derivative(material.electrical_conductivity,T), cold_temp, hot_temp)[0]
    la_ratio = (sigTL * minimum_Q + inte)/(current**2)

    return la_ratio

# ----------------------------------------------------------------------
#  calc_current
# ----------------------------------------------------------------------
def calc_current(Cryogenic_Lead, current):

    """Estimates the heat flow into the cryogenic environment when a current other than the current the lead was optimised for is flowing. Assumes the temperature difference remains constant.

    Assumptions:
    N/A

    Source:
    N/A

    Inputs:
    Cryogenic_Lead. 
        outputs.optimum_current     [A]
        outputs.minimum_Q           [W]
        outputs.unpowered_Q         [W]
        cold_temp                   [K]
        hot_temp                    [K]
        outputs.cross_section       [m2]
        length                      [m]
        material

    Outputs:
        [lead cooling power, cryogenic loading due to lead]     [w,w]

    Properties Used:
    N/A
    """  

    c_l = Cryogenic_Lead

    design_current      = c_l.outputs.optimum_current
    design_Q            = c_l.outputs.minimum_Q
    zero_Q              = c_l.outputs.unpowered_Q
    cold_temp           = c_l.cold_temp
    hot_temp            = c_l.hot_temp
    cs_area             = c_l.outputs.cross_section
    length              = c_l.length
    material            = c_l.material

    # The thermal gradient along the lead is assumed to remain constant for all currents below the design current. The resistance remains constant if the temperature remains constant. The estimated heat flow is reduced in proportion with the carried current.
    if current <= design_current:
        proportion      = current/design_current
        R               = design_Q/(design_current**2.0)
        power           = R*current**2.0
        Q               = zero_Q + proportion * (design_Q - zero_Q)

    # If the supplied current is higher than the design current the maximum temperature in the lead will be higher than ambient. Solve by dividing the lead at the maximum temperature point.
    else:
        # Initial guess at max temp in lead
        max_temp        = 2 * hot_temp
        # Find actual maximum temperature by bisection, accept result within 1% of correct.
        error           = 1
        guess_over      = 0
        guess_diff      = hot_temp

        while error > 0.01:
            # Find length of warmer part of lead
            warm_Q          = compute_minimum_power(material, hot_temp, max_temp, current)

            warm_la         = compute_optimal_ratio(material, hot_temp, max_temp, current, warm_Q)
            warm_length     = cs_area * warm_la
            # Find length of cooler part of lead
            cool_Q          = compute_minimum_power(material, cold_temp, max_temp, current)
            cool_la         = compute_optimal_ratio(material, cold_temp, max_temp, current, cool_Q)
            cool_length     = cs_area * cool_la
            # compare lead length with known lead length as test of the max temp guess
            test_length     = warm_length + cool_length
            error           = abs((test_length-length)/length)
            # change the guessed max_temp
            # A max_temp too low will result in the test length being too long
            if test_length > length:
                if guess_over == 0:             # query whether solving by bisection yet
                    guess_diff  = max_temp      # if not, continue to double guess
                    max_temp    = 2*max_temp
                else:
                    max_temp    = max_temp + guess_diff
            else:
                guess_over  = 1              # set flag that bisection range found
                max_temp    = max_temp - guess_diff
            # Prepare guess difference for next iteration
            guess_diff  = 0.5*guess_diff
            # The cool_Q is the cryogenic heat load as warm_Q is sunk to ambient
            Q           = cool_Q
            # All Q is out of the lead, so the electrical power use in the lead is the sum of the Qs
            power       = warm_Q + cool_Q

    return [Q,power]


# RCAIDE/Framework/Analyses/Noise/Frequency_Domain_Buildup.py
# 
# 
# Created:   Jul 2023, M. Clarke
# Modified:  Oct 2024, A. Molloy

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------  
# RCAIDE Imports
import RCAIDE
from RCAIDE.Library.Methods.Noise.Common.decibel_arithmetic                           import SPL_arithmetic  
from RCAIDE.Library.Methods.Noise.Common.generate_hemisphere_microphone_locations     import generate_hemisphere_microphone_locations  
from RCAIDE.Library.Methods.Noise.Frequency_Domain_Buildup.Rotor.compute_rotor_noise  import compute_rotor_noise 
from .Noise      import Noise

# package imports
import numpy as np 

# ----------------------------------------------------------------------------------------------------------------------
#  Frequency_Domain_Buildup
# ----------------------------------------------------------------------------------------------------------------------
class Frequency_Domain_Buildup(Noise):
    """This is an acoustic analysis based on a collection of frequency domain methods 

     Assumptions: 
 
     Source:
     N/A
 
     Inputs:
     None
 
     Outputs:
     None
 
     Properties Used:
     N/A 
    """    
    
    def __defaults__(self):
        
        """ This sets the default values for the analysis.
        
            Assumptions:
            Ground microphone angles start in front of the aircraft (0 deg) and sweep in a lateral direction 
            to the starboard wing and around to the tail (180 deg)
            
            Source:
            N/A
            
            Inputs:
            None
            
            Output:
            None
            
            Properties Used:
            N/A
        """
        
        # Initialize quantities 
        self.tag                                   =  "Frequency_Domain_Buildup"        
        self.settings.fidelity                     = 'line_source'
        self.settings.use_plane_loading_surrogate =  True 
    def evaluate_noise(self,segment):
        """ Process vehicle to setup vehicle, condititon and configuration
    
        Assumptions:
        None
    
        Source:
        N/4
    
        Inputs:
        self.settings.
            center_frequencies  - 1/3 octave band frequencies   [unitless]
    
        Outputs:
        None
    
        Properties Used:
        self.vehicle
        """         
    
        # unpack 
        config        = segment.analyses.noise.vehicle 
        settings      = self.settings  
        conditions    = segment.state.conditions  
        dim_cf        = len(settings.center_frequencies ) 
        ctrl_pts      = int(segment.state.numerics.number_of_control_points) 
         
        microphone_locations =  generate_hemisphere_microphone_locations(settings)     
        N_hemisphere_mics    =  len(microphone_locations)
        
        # create empty arrays for results      
        total_SPL_dBA          = np.ones((ctrl_pts,N_hemisphere_mics))*1E-16 
        total_SPL_spectra      = np.ones((ctrl_pts,N_hemisphere_mics,dim_cf))*1E-16  
         
        # iterate through sources and iteratively add rotor noise
        rotor_tag = None
        i = 0
        for network in config.networks:
            for propulsor in network.propulsors:
                for sub_tag , sub_item in  propulsor.items():
                    if isinstance(sub_item, RCAIDE.Library.Components.Propulsors.Converters.Rotor): 
                        rotor_tag         = compute_rotor_noise(microphone_locations,propulsor,sub_item,segment,settings, rotor_index = i, previous_rotor_tag= rotor_tag, identical_propulsors=network.identical_propulsors)   
                        total_SPL_dBA     = SPL_arithmetic(np.concatenate((total_SPL_dBA[:,None,:],conditions.noise[propulsor.tag][sub_item.tag].SPL_dBA[:,None,:]),axis =1),sum_axis=1)
                        total_SPL_spectra = SPL_arithmetic(np.concatenate((total_SPL_spectra[:,None,:,:],conditions.noise[propulsor.tag][sub_item.tag].SPL_1_3_spectrum[:,None,:,:]),axis =1),sum_axis=1) 
                        i += 1
                        
        conditions.noise.hemisphere_SPL_dBA              = total_SPL_dBA
        conditions.noise.hemisphere_SPL_1_3_spectrum_dBA = total_SPL_spectra  
        return


# RCAIDE/Library/Components/Propulsors/Electric_Rotor.py
#  
# 
# Created:  Mar 2024, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 
 # RCAIDE imports 
from .   import Propulsor 
from RCAIDE.Library.Methods.Propulsors.Electric_Rotor_Propulsor.unpack_electric_rotor_unknowns             import unpack_electric_rotor_unknowns
from RCAIDE.Library.Methods.Propulsors.Electric_Rotor_Propulsor.pack_electric_rotor_residuals              import pack_electric_rotor_residuals
from RCAIDE.Library.Methods.Propulsors.Electric_Rotor_Propulsor.append_electric_rotor_conditions           import append_electric_rotor_conditions
from RCAIDE.Library.Methods.Propulsors.Electric_Rotor_Propulsor.compute_electric_rotor_performance         import compute_electric_rotor_performance, reuse_stored_electric_rotor_data
from RCAIDE.Library.Methods.Propulsors.Electric_Rotor_Propulsor.append_electric_rotor_residual_and_unknown import append_electric_rotor_residual_and_unknown

# ---------------------------------------------------------------------------------------------------------------------- 
#  Electric_Rotor
# ----------------------------------------------------------------------------------------------------------------------  
class Electric_Rotor(Propulsor):
    """
    A propulsion system class that combines an electric motor with a rotor for vertical and forward flight.
    
    Attributes
    ----------
    tag : str
        Identifier for the propulsion system, defaults to 'electric_rotor'
    
    motor : None or Motor
        The electric motor component that provides rotational power
        
    rotor : None or Rotor
        The rotor component that generates lift and thrust
        
    electronic_speed_controller : None or ESC
        The electronic speed controller that regulates power to the motor
    
    Notes
    -----
    This class models an electric propulsion system where an electric motor drives 
    a rotor to generate lift and thrust. The system includes an electronic speed 
    controller (ESC) to regulate power delivery from the electrical system to the motor.
    
    The class provides methods for:
    - Computing rotor performance (thrust, moment, power)
    - Managing operating conditions
    - Handling system states and residuals
    - Reusing stored performance data for computational efficiency
    
    This propulsor type is commonly used in:
    - Multicopters
    - eVTOL aircraft
    - Hybrid helicopters
    
    **Definitions**

    'Rotor'
        A rotating assembly of airfoils (blades) that generates lift and thrust 
        through aerodynamic forces
        
    'Electronic Speed Controller (ESC)'
        Device that controls the speed of the electric motor by regulating 
        power delivery based on input commands
    
    See Also
    --------
    RCAIDE.Library.Components.Propulsors.Propulsor
    RCAIDE.Library.Components.Energy
    """
    def __defaults__(self):    
        # setting the default values
        self.tag                          = 'electric_rotor'    
        self.motor                        = None
        self.rotor                        = None 
        self.electronic_speed_controller  = None 

    def append_operating_conditions(self,segment):
        """
        Appends operating conditions to the segment.
        """
        append_electric_rotor_conditions(self,segment)
        return
    
    def append_propulsor_unknowns_and_residuals(self,segment):
        """
        Appends propulsor unknowns and residuals to the segment.
        """
        append_electric_rotor_residual_and_unknown(self,segment)
        return

    def unpack_propulsor_unknowns(self,segment):  
        """
        Unpacks propulsor unknowns from the segment.
        """
        unpack_electric_rotor_unknowns(self,segment)
        return 

    def pack_propulsor_residuals(self,segment): 
        """
        Packs propulsor residuals into the segment.
        """
        pack_electric_rotor_residuals(self,segment)
        return    
    
    def compute_performance(self,state,voltage,center_of_gravity = [[0, 0, 0]]):
        """
        Computes propulsor performance including thrust, moment, and power. 
        """
        thrust,moment,power,stored_results_flag,stored_propulsor_tag =  compute_electric_rotor_performance(self,state,voltage,center_of_gravity)
        return thrust,moment,power,stored_results_flag,stored_propulsor_tag
    
    def reuse_stored_data(electric_rotor,state,network,stored_propulsor_tag,center_of_gravity = [[0, 0, 0]]):
        """
        Reuses stored propulsor data for performance calculations.
        """
        thrust,moment,power = reuse_stored_electric_rotor_data(electric_rotor,state,network,stored_propulsor_tag,center_of_gravity)
        return thrust,moment,power


# RCAIDE/Library/Methods/Energy/Propulsors/Turboshaft_Propulsor/compute_power.py
# 
# 
# Created:  Jul 2023, M. Clarke
# Modified: Jun 2024, M. Guidotti & D.J. Lee

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 
 
# Python package imports
import numpy                               as np

# ----------------------------------------------------------------------------------------------------------------------
#  compute_power
# ----------------------------------------------------------------------------------------------------------------------

def compute_power(turboshaft,turboshaft_conditions,conditions):
    """Computes power and other properties as below.

    Assumptions:
    Perfect gas
    Turboshaft engine with free power turbine

    Sources:
    [1] https://soaneemrana.org/onewebmedia/ELEMENTS%20OF%20GAS%20TURBINE%20PROPULTION2.pdf - Page 332 - 336
    [2] https://www.colorado.edu/faculty/kantha/sites/default/files/attached-files/70652-116619_-_luke_stuyvenberg_-_dec_17_2015_1258_pm_-_stuyvenberg_helicopterturboshafts.pdf
    
    Inputs:
    conditions.freestream.
      isentropic_expansion_factor              [-] (gamma)
      specific_heat_at_constant_pressure       [J/(kg K)]
      velocity                                 [m/s]
      speed_of_sound                           [m/s]
      mach_number                              [-]
      pressure                                 [Pa]
      gravity                                  [m/s^2]
    conditions.throttle                        [-] (.1 is 10%)
    turboshaft.inputs.                         
      fuel_to_air_ratio                        [-]
      total_temperature_reference              [K]
      total_pressure_reference                 [Pa]
      core_nozzle.                             
        velocity                               [m/s]
        static_pressure                        [Pa]
        area_ratio                             [-]
      fan_nozzle.                              
        velocity                               [m/s]
        static_pressure                        [Pa]
        area_ratio                             [-]
      number_of_engines                        [-]
      bypass_ratio                             [-]
      flow_through_core                        [-] percentage of total flow (.1 is 10%)
      flow_through_fan                         [-] percentage of total flow (.1 is 10
    combustor.outputs.stagnation_temperature   [K]
    compressor.pressure_ratio                  [-]
    turboshaft.conversion_efficiency           [-]
                                               
    Outputs:                                   
    turboshaft.outputs.                        
      thrust                                   [N]
      non_dimensional_thrust                   [-]
      core_mass_flow_rate                      [kg/s]
      fuel_flow_rate                           [kg/s]
      power                                    [W]
      power_specific_fuel_consumption          [kg/(W*s)]
      Specific Impulse                         [s]
                                               
    Properties Used:                           
    turboshaft.                                
      reference_temperature                    [K]
      reference_pressure                       [Pa]
      compressor_nondimensional_massflow       [-]
      SFC_adjustment                           [-]
    """           
    #unpack the values
    fuel_type                                  = turboshaft.fuel_type
    LHV                                        = fuel_type.lower_heating_value                                                                        
    gamma                                      = conditions.freestream.isentropic_expansion_factor                                                      
    a0                                         = conditions.freestream.speed_of_sound                                                                   
    M0                                         = conditions.freestream.mach_number                                                                      
    Cp                                         = conditions.freestream.Cp                                                                               # Source [2]
    total_temperature_reference                = turboshaft_conditions.total_temperature_reference                                                          
    total_pressure_reference                   = turboshaft_conditions.total_pressure_reference                                                             # Source [1]
    eta_c                                      = turboshaft.conversion_efficiency                                                                       # Source [2]
                                                                                                                                                        
    #unpacking from turboshaft                                                                                                                          
    Tref                                       = turboshaft.reference_temperature                                                                       # Source [1]
    Pref                                       = turboshaft.reference_pressure                                                                          # Source [1]
    Tt4                                        = turboshaft_conditions.combustor_stagnation_temperature                                                    
    pi_c                                       = turboshaft.compressor.pressure_ratio                                                                   
    m_dot_compressor                           = turboshaft.compressor.mass_flow_rate                                                                   # Source [2]
                                                                                                                                                        
    tau_lambda                                 = Tt4/total_temperature_reference                                                                        
    tau_r                                      = 1 + ((gamma - 1)/2)*M0**2                                                                              
    tau_c                                      = pi_c**((gamma - 1)/gamma)                                                                              
    tau_t                                      = (1/(tau_r*tau_c)) + ((gamma - 1)*M0**2)/(2*tau_lambda*eta_c**2)                                        # Source [2]
    #tau_t                                      = x/(tau_r*tau_c)                                                                                      # Source [1]
    tau_tH                                     = 1 - (tau_r/tau_lambda)*(tau_c - 1)                                                                    # Source [2]
    tau_tL                                     = tau_t/tau_tH                                                                                          # Source [2]
    #x                                          = 1.02                                                                                                  # Source [1] Page 335
    x                                          = tau_t*tau_r*tau_c                                                                                     # Source [1] 
    #C_shaft                                    = tau_lambda*(1 - x/(tau_r*tau_c)) - tau_r*(tau_c - 1)                                                  # Source [1]
    C_shaft                                    = tau_lambda*(1 - tau_t) - tau_r*(tau_c - 1)                                                             # Source [1]    

    #Computing Specifc Thrust
    Tsp                                        = a0*(((2/(gamma - 1))*(tau_lambda/(tau_r*tau_c))*(tau_r*tau_c*tau_t - 1))**eta_c - M0)                  # Source [2]
    
    #computing the core mass flow              
    m_dot_air                                  = m_dot_compressor*np.sqrt(Tref/total_temperature_reference)*(total_pressure_reference/Pref)             # Source [1]
    
    #Computing Specifc Power
    #Psp                                        = Cp*total_temperature_reference*C_shaft                                                                 # Source [1] 
    Psp                                        =  Cp*total_temperature_reference*tau_lambda*tau_tH*(1 - tau_tL)*eta_c                                   # Source [2]    
    
    #Computing Power 
    Power                                      = Psp*m_dot_air                                                                                          
    #Power                                      = m_dot_air*Cp*total_temperature_reference*(tau_lambda*(1 - tau_t) - tau_r*(tau_c - 1))                 # Source [2]

    #fuel to air ratio
    f                                          = (Cp*total_temperature_reference/LHV)*(tau_lambda - tau_r*tau_c)                                        # Source [2]    
                                                                                                                                               
    #fuel flow rate                             
    #fuel_flow_rate                             = Power*PSFC*1./Units.hour                                                                              # Source [1]
    fuel_flow_rate                             = f*m_dot_air
    
    #Computing the PSFC                        
    PSFC                                       = f/Psp                                                                                                  
    #PSFC                                       = (tau_lambda/(C_shaft*LHV))                                                                            # Source [1]  
    
    #Computing the thermal efficiency                       
    eta_T                                      = 1 - (tau_r*(tau_c - 1))/(tau_lambda*(1 - x/(tau_r*tau_c)))                                             # Source [1]    

    #pack outputs
    turboshaft_conditions.power_specific_fuel_consumption   = PSFC
    turboshaft_conditions.fuel_flow_rate                    = fuel_flow_rate                                                                              
    turboshaft_conditions.power                             = Power
    turboshaft_conditions.non_dimensional_power             = Psp
    turboshaft_conditions.non_dimensional_thrust            = Tsp
    turboshaft_conditions.thermal_efficiency                = eta_T

    return 



## @ingroup Analyses-Planets
# Planet.py
#
# Created:  
# Modified: Feb 2016, Andrew Wendorff

# ----------------------------------------------------------------------
#  Imports
# ----------------------------------------------------------------------
import RCAIDE
from RCAIDE.Framework.Core import Data 
from RCAIDE.Framework.Analyses import Analysis

# ----------------------------------------------------------------------
#  Analysis
# ----------------------------------------------------------------------

## @ingroup Analyses-Planets
class Planet(Analysis):
    """ RCAIDE.Framework.Analyses.Planet()
    """
    
    def __defaults__(self):
        
        """This sets the default values and methods for the analysis.
    
            Assumptions:
            Planet is Earth.
    
            Source:
            N/A
    
            Inputs:
            None
    
            Outputs:
            None
    
            Properties Used:
            N/A
            """                  
        
        
        self.tag      = 'planet'
        self.settings = Data()
        
        
        


# RCAIDE/Library/Compoments/Energy/Networks/Distribution/__init__.py
# 

"""
Energy sources module providing components for aircraft power generation and storage

This module contains implementations for various energy source components including
batteries, fuel tanks, and other power generation systems. These components serve
as the primary energy providers in aircraft propulsion systems.

See Also
--------
RCAIDE.Library.Components.Energy.Modulators
    Energy control systems
RCAIDE.Library.Components.Energy.Sources
    Fuel storage and delivery systems
"""

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
 
from .Electrical_Bus                       import Electrical_Bus
from .Fuel_Line                            import Fuel_Line
from .Coolant_Line                         import Coolant_Line




# RCAIDE/Library/Methods/Weights/Correlation_Buildups/General_Aviation/compute_main_wing_weight.py
# 
# 
# Created:  Sep 2024, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

# RCAIDE 
from RCAIDE.Framework.Core import  Units
import  numpy as  np

# ----------------------------------------------------------------------------------------------------------------------
# Main Wing Weight 
# ----------------------------------------------------------------------------------------------------------------------
def compute_main_wing_weight(S_wing, m_fuel, AR_w, sweep_w, q_c, taper_w, t_c_w,Nult,TOW):      
    """ 
        Calculate the weight of the main wing of an aircraft     

        Source:
            Raymer, Aircraft Design: A Conceptual Approach (pg 460 in 4th edition)
        
        Inputs:
            S_wing- area of the main wing                                              [meters**2]
            m_fuel - predicted weight of fuel in the wing                              [kilograms]
            AR_w -aspect ratio of main wing                                            [dimensionless]
            sweep_w - quarter chord sweep of the main wing                             [radians]
            q_c - dynamic pressure at cruise                                           [Pascals]
            taper_w - taper ratio of wing                                              [dimensionless]
            t_c_w -thickness to chord ratio of wing                                    [dimensionless]
            Nult - ultimate load of the aircraft                                       [dimensionless]
            TOW - maximum takeoff weight of the aircraft                               [kilograms]
   
        Outputs:
            output - a dictionary with outputs:
                W_main_wing - weight of the vertical fin portion of the vertical tail [kilograms]
    """     
    # unpack inputs 
    W_0  = TOW / Units.lb # Convert kg to lbs
    S_w  = S_wing/ (Units.ft**2) # Convert from meters squared to ft squared  
    W_fw = m_fuel/Units.lbs #convert from kg to lbs
    q    = q_c /(Units.lbs/(Units.ft**2.))

    # Calculate weight of wing for traditional aircraft vertical tail without rudder
    weight_English = .036 * (S_w**.758)*(W_fw**.0035)*((AR_w/(np.cos(sweep_w)**2))**.6)*(q**.006)*(taper_w**.04)*((100.*t_c_w/np.cos(sweep_w))**(-.3))*((Nult*W_0)**.49)
    
    # packup outputs    
    weight =  weight_English * Units.lbs # Convert from lbs to kg

    return weight



# RCAIDE/Library/Missions/Segments/Descent/Constant_Speed_Constant_Angle.py
# 
# 
# Created:  Jul 2023, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------  
#  IMPORT 
# ----------------------------------------------------------------------------------------------------------------------  
# package imports 
import numpy as np

# ----------------------------------------------------------------------------------------------------------------------  
#  Initialize Conditions
# ----------------------------------------------------------------------------------------------------------------------  
def initialize_conditions(segment):
    """Sets the specified conditions which are given for the segment type.

    Assumptions:
    Constant speed and constant descent angle

    Source:
    N/A

    Inputs:
    segment.descent_angle                               [radians]
    segment.altitude_start                              [meters]
    segment.altitude_end                                [meters]
    segment.air_speed                                   [meters/second]
    segment.state.numerics.dimensionless.control_points [array]

    Outputs:
    conditions.frames.inertial.velocity_vector  [meters/second]
    conditions.frames.inertial.position_vector  [meters]
    conditions.freestream.altitude              [meters]
    conditions.frames.inertial.time             [seconds]

    Properties Used:
    N/A
    """        
    
    # unpack
    descent_angle= segment.descent_angle
    air_speed    = segment.air_speed   
    alt0         = segment.altitude_start 
    altf         = segment.altitude_end 
    beta         = segment.sideslip_angle
    t_nondim     = segment.state.numerics.dimensionless.control_points
    conditions   = segment.state.conditions  

    # check for initial altitude
    if alt0 is None:
        if not segment.state.initials: raise AttributeError('initial altitude not set')
        alt0 = -1.0 * segment.state.initials.conditions.frames.inertial.position_vector[-1,2]
    
    # check for initial velocity vector
    if air_speed is None:
        if not segment.state.initials: raise AttributeError('initial airspeed not set')
        air_speed  =  np.linalg.norm(segment.state.initials.conditions.frames.inertial.velocity_vector[-1,:])     
            
    # discretize on altitude
    alt = t_nondim * (altf-alt0) + alt0
    
    # process velocity vector
    v_mag = air_speed
    v_x   = np.cos(beta)* v_mag * np.cos(-descent_angle)
    v_y   = np.sin(beta)* v_mag * np.cos(-descent_angle)
    v_z   = -v_mag * np.sin(-descent_angle)
    
    # pack conditions    
    conditions.frames.inertial.velocity_vector[:,0] = v_x
    conditions.frames.inertial.velocity_vector[:,1] = v_y
    conditions.frames.inertial.velocity_vector[:,2] = v_z
    conditions.frames.inertial.position_vector[:,2] = -alt[:,0] # z points down
    conditions.freestream.altitude[:,0]             =  alt[:,0] # positive altitude in this context


# RCAIDE/Library/Missions/Common/compute_point_to_point_geospacial_data.py
# 
# 
# Created:  Jul 2023, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------  
import RCAIDE
from RCAIDE.Framework.Core import Units 
from scipy.interpolate import griddata
import numpy as np

# ----------------------------------------------------------------------
#  Compute Point to Point Geospacial Data
# --------------------------------------------------------------------- 
def compute_point_to_point_geospacial_data(settings):
    """This computes the absolute microphone/observer locations on a defined topography
            
    Assumptions: 
        topography_file is a text file obtained from https://topex.ucsd.edu/cgi-bin/get_data.cgi
    
    Source:
        N/A  

    Inputs:   
        topography_file                        - file of lattide, longitude and elevation points     
        origin_coordinates                     - coordinates of origin location                                              [degrees]
        destination_coordinates                - coordinates of destimation location                                            [degrees]  
        
    Outputs:                                   
        latitude_longitude_micrphone_locations - latitude-longitude and elevation coordinates of all microphones in domain      [deg,deg,m]  
        flight_range                           - gound distance between origin and destination location                      [meters]              
        true_course                            - true course angle measured clockwise from true north                     [radians]                      
        origin_location                        - cartesial coordinates of origin location relative to computational domain   [meters]                   
        destination_xyz_location               - cartesial coordinates of destination location relative to computational domain [meters]    
    
    Properties Used:
        N/A       
    """     
    # convert cooordinates to array 
    origin_coordinates   = np.asarray(settings.aircraft_origin_coordinates)
    destination_coordinates = np.asarray(settings.aircraft_destination_coordinates)
    
    # extract data from file 
    data  = np.loadtxt(settings.topography_file)
    Long  = data[:,0]
    Lat   = data[:,1]
    Elev  = data[:,2] 

    x_min_coord = np.min(Lat)
    y_min_coord = np.min(Long)
    dep_lat     = origin_coordinates[0]
    dep_long    = origin_coordinates[1]
    des_lat     = destination_coordinates[0]
    des_long    = destination_coordinates[1]
    if dep_long < 0: 
        dep_long = 360 + dep_long
    if des_long< 0:
        des_long =360 +  des_long 
    
    bottom_left_map_coords   = np.array([x_min_coord,y_min_coord])  
    x0_coord                 = np.array([dep_lat,y_min_coord])
    y0_coord                 = np.array([x_min_coord,dep_long])
    x1_coord                 = np.array([des_lat,y_min_coord])
    y1_coord                 = np.array([x_min_coord,des_long])  
    
    x0 = RCAIDE.Framework.Analyses.Geodesics.Geodesics.Calculate_Distance(x0_coord,bottom_left_map_coords) * Units.kilometers
    y0 = RCAIDE.Framework.Analyses.Geodesics.Geodesics.Calculate_Distance(y0_coord,bottom_left_map_coords) * Units.kilometers
    x1 = RCAIDE.Framework.Analyses.Geodesics.Geodesics.Calculate_Distance(x1_coord,bottom_left_map_coords) * Units.kilometers
    y1 = RCAIDE.Framework.Analyses.Geodesics.Geodesics.Calculate_Distance(y1_coord,bottom_left_map_coords) * Units.kilometers
    
    lat_flag             = np.where(origin_coordinates<0)[0]
    origin_coordinates[lat_flag]  = origin_coordinates[lat_flag] + 360 
    long_flag            = np.where(destination_coordinates<0)[0]
    destination_coordinates[long_flag] = destination_coordinates[long_flag] + 360 
    z0                   = griddata((Lat,Long), Elev, (np.array([origin_coordinates[0]]),np.array([origin_coordinates[1]])), method='nearest')[0]
    z1                   = griddata((Lat,Long), Elev, (np.array([destination_coordinates[0]]),np.array([destination_coordinates[1]])), method='nearest')[0] 
    dep_loc              = np.array([x0,y0,z0])
    des_loc              = np.array([x1,y1,z1])
    
    # pack data 
    settings.aircraft_origin_location      = dep_loc
    settings.aircraft_destination_location = des_loc 
        
    return 



# RCAIDE/Library/Methods/Propulsors/Electric_Rotor_Propulsor/pack_ice_propeller_residuals.py
# (c) Copyright 2023 Aerospace Research Community LLC
# 
# Created:  Jun 2024, M. Clarke   

# ---------------------------------------------------------------------------------------------------------------------- 
#  pack ice propeller residuals
# ----------------------------------------------------------------------------------------------------------------------  

def pack_ice_propeller_residuals(propulsor,segment):  
    engine             = propulsor.engine
    propeller          = propulsor.propeller  
    q_engine           = segment.state.conditions.energy[propulsor.tag][engine.tag].torque
    q_prop             = segment.state.conditions.energy[propulsor.tag][propeller.tag].torque 
    segment.state.residuals.network[propulsor.tag + '_rotor_engine_torque'] = q_engine - q_prop 
    return 





# ----------------------------------------------------------------------
#  Imports
# ----------------------------------------------------------------------
import numpy as np

# ----------------------------------------------------------------------
#  Compressible Turbulent Flat Plate
# ----------------------------------------------------------------------

## @ingroup Methods-Aerodynamics-Common-Fidelity_Zero-Helper_Functions
def compressible_turbulent_flat_plate(Re,Ma,Tc):
    """Computes the coefficient of friction for a flat plate given the 
    input parameters. Also returns the correction terms used in the
    computation.

    Assumptions:
    Reynolds number between 10e5 and 10e9
    Fully turbulent

    Source:
    adg.stanford.edu (Stanford AA241 A/B Course Notes)

    Inputs:
    Re (Reynolds number)                                             [Unitless]
    Ma (Mach number)                                                 [Unitless]
    Tc (temperature)                                                 [K]

    Outputs:
    cf_comp (coefficient of friction)                                [Unitless]
    k_comp (compressibility correction)                              [Unitless]
    k_reyn (Reynolds number correction)                              [Unitless]

    Properties Used:
    N/A
    """     

    # incompressible skin friction coefficient
    cf_inc = 0.455/(np.log10(Re))**2.58
    
    # compressibility correction
    Tw = Tc * (1. + 0.178*Ma**2.)
    Td = Tc * (1. + 0.035*Ma**2. + 0.45*(Tw/Tc - 1.))
    k_comp = (Tc/Td) 
    
    # reynolds correction
    Rd_w   = Re * (Td/Tc)**1.5 * ( (Td+216.) / (Tc+216.) )
    k_reyn = (Re/Rd_w)**0.2
    
    # apply corrections
    cf_comp = cf_inc * k_comp * k_reyn
    
    return cf_comp, k_comp, k_reyn

  
# ----------------------------------------------------------------------
#   Module Tests
# ----------------------------------------------------------------------
# this will run from command line, put simple tests for your code here
if __name__ == '__main__':    
    
    (cf_comp, k_comp, k_reyn) = compressible_turbulent_flat_plate(1.0*10.0**7.0,0.0,216.0)
    
    print(cf_comp)
    print(k_comp)
    print(k_reyn)    



# RCAIDE/Library/Missions/Common/Update/linear_inertial_horizontal_position.py
# 
# 
# Created:  Jul 2023, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

# Package imports 
import numpy as np

# ----------------------------------------------------------------------------------------------------------------------
#  Integrate Position
# ----------------------------------------------------------------------------------------------------------------------
 
def linear_inertial_horizontal_position(segment):
    """ Determines how far the airplane has traveled. 
    
        Assumptions:
        Assumes a flat earth, this is planar motion.
        
        Inputs:
            segment.state.conditions:
                frames.inertial.position_vector [meters]
                frames.inertial.velocity_vector [meters/second]
            segment.state.numerics.time.integrate       [float]
            
        Outputs:
            segment.state.conditions:           
                frames.inertial.position_vector [meters]

        Properties Used:
        N/A
                                
    """        

    conditions = segment.state.conditions
    psi        = segment.true_course       # sign convetion is clockwise positive
    cpts       = int(segment.state.numerics.number_of_control_points)
    x0         = conditions.frames.inertial.position_vector[0,None,0:1+1]
    R0         = conditions.frames.inertial.aircraft_range[0,None,0:1+1]
    vx         = conditions.frames.inertial.velocity_vector[:,0:1+1]
    I          = segment.state.numerics.time.integrate  
    trajectory = np.repeat( np.atleast_2d(np.array([np.cos(psi),np.sin(psi)])),cpts , axis = 0) 
    
    # integrate
    x = np.dot(I,vx)  
    x[:,1] = x[:,0]
    
    # pack
    conditions.frames.inertial.position_vector[:,0:1+1] = x0 + x[:,:]*trajectory
    conditions.frames.inertial.aircraft_range[:,0]      = R0 + x[:,0]  
    
    return


# RCAIDE/Library/Plots/Geometry/plot_3d_vehicle.py
# 
# 
# Created:  Jul 2023, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 
from RCAIDE.Library.Plots.Geometry.plot_3d_fuselage             import plot_3d_fuselage
from RCAIDE.Library.Plots.Geometry.plot_3d_wing                 import plot_3d_wing 
from RCAIDE.Library.Plots.Geometry.plot_3d_nacelle              import plot_3d_nacelle
from RCAIDE.Library.Plots.Geometry.plot_3d_rotor                import plot_3d_rotor

# python imports 
import numpy as np 
import plotly.graph_objects as go  

# ----------------------------------------------------------------------------------------------------------------------
#  PLOTS
# ----------------------------------------------------------------------------------------------------------------------  
def plot_3d_vehicle(vehicle,
                    show_axis                   = False,
                    save_figure                 = False,
                    save_filename               = "Vehicle_Geometry",
                    alpha                       = 1.0,  
                    min_x_axis_limit            =  -5,
                    max_x_axis_limit            =  40,
                    min_y_axis_limit            =  -20,
                    max_y_axis_limit            =  20,
                    min_z_axis_limit            =  -20,
                    max_z_axis_limit            =  20, 
                    camera_eye_x                = -1.5,
                    camera_eye_y                = -1.5,
                    camera_eye_z                = .8,
                    camera_center_x             = 0.,
                    camera_center_y             = 0.,
                    camera_center_z             = -0.5,
                    show_figure                 = True):
    """This plots a 3D representation of the aircraft 

    Assumptions:
    None

    Source:
    None

    Inputs:
       vehicle                      - vehicle data structure 
       show_axis                    - plot axis flag          
       save_figure                  - safe figure flag              
       alpha                        - opacity                   
       show_wing_control_points     - show control point flag  
       show_rotor_wake_vortex_core  - show rotor wake flag 
       save_filename                - filename for saving  
       x_axis_limit                 - limits of axis  
       y_axis_limit                 - limits of axis  
       z_axis_limit                 - limits of axis  
       show_figure                  - show figure flag  

    Outputs:
    Plots

    Properties Used:
    N/A
    """

    print("\nPlotting vehicle") 
    camera        =  dict(eye=dict(x=camera_eye_x, y=camera_eye_y, z=camera_eye_z), center=dict(x= camera_center_x, y=camera_center_x, z= camera_center_z))   
    plot_data     = []
    
    plot_data,x_min,x_max,y_min,y_max,z_min,z_max  = generate_3d_vehicle_geometry_data(plot_data,
                                                                                       vehicle,
                                                                                       alpha,  
                                                                                       min_x_axis_limit,
                                                                                       max_x_axis_limit,
                                                                                       min_y_axis_limit,
                                                                                       max_y_axis_limit,
                                                                                       min_z_axis_limit,
                                                                                       max_z_axis_limit)
    

    fig = go.Figure(data=plot_data)
    fig.update_scenes(aspectmode   = 'cube',
                      xaxis_visible=show_axis,
                      yaxis_visible=show_axis,
                      zaxis_visible=show_axis
                      )
    fig.update_layout( 
             width     = 1500,
             height    = 1500, 
             scene = dict(
                        xaxis = dict(backgroundcolor="grey", gridcolor="white", showbackground=show_axis,
                                     zerolinecolor="white", range=[x_min,x_max]),
                        yaxis = dict(backgroundcolor="grey", gridcolor="white", showbackground=show_axis, 
                                     zerolinecolor="white", range=[y_min,y_max]),
                        zaxis = dict(backgroundcolor="grey",gridcolor="white",showbackground=show_axis,
                                     zerolinecolor="white", range=[z_min,z_max])),             
             scene_camera=camera) 
    fig.update_coloraxes(showscale=False) 
    fig.update_traces(opacity = alpha)
    if save_figure:
        fig.write_image(save_filename + ".png")
        
    if show_figure:
        fig.write_html( save_filename + '.html', auto_open=True) 
    
    return     

def generate_3d_vehicle_geometry_data(plot_data,
                                      vehicle, 
                                      alpha                       = 1.0,  
                                      min_x_axis_limit            =  -5,
                                      max_x_axis_limit            =  40,
                                      min_y_axis_limit            =  -20,
                                      max_y_axis_limit            =  20,
                                      min_z_axis_limit            =  -20,
                                      max_z_axis_limit            =  20, ):
    """ This plots the 3D surface of the network

    Assumptions:
    None

    Source:
    None

    Inputs:
       vehicle                      - vehicle data structure          
       alpha                        - opacity  
       save_filename                - filename for saving  
       x_axis_limit                 - limits of axis  
       y_axis_limit                 - limits of axis  
       z_axis_limit                 - limits of axis   

    Properties Used:
    N/A
    """ 
    
    # -------------------------------------------------------------------------
    # PLOT WING
    # ------------------------------------------------------------------------- 
    number_of_airfoil_points = 21
    for wing in vehicle.wings:
        plot_data       = plot_3d_wing(plot_data,wing,number_of_airfoil_points , color_map='greys',alpha=1) 
        
    # -------------------------------------------------------------------------
    # PLOT FUSELAGE
    # ------------------------------------------------------------------------- 
    for fus in vehicle.fuselages:
        plot_data = plot_3d_fuselage(plot_data,fus,color_map = 'teal')

    
    # -------------------------------------------------------------------------
    # PLOT BOOMS
    # ------------------------------------------------------------------------- 
    for boom in vehicle.booms:
        plot_data = plot_3d_fuselage(plot_data,boom,color_map = 'gray') 
        
    # -------------------------------------------------------------------------
    # PLOT ROTORS
    # ------------------------------------------------------------------------- 
    number_of_airfoil_points = 11
    for network in vehicle.networks:
        plot_data = plot_3d_energy_network(plot_data,network,number_of_airfoil_points,color_map = 'turbid' )
 
    return plot_data,min_x_axis_limit,max_x_axis_limit,min_y_axis_limit,max_y_axis_limit,min_z_axis_limit,max_z_axis_limit

def plot_3d_energy_network(plot_data,network,number_of_airfoil_points,color_map):
    """ This plots the 3D surface of the network

    Assumptions:
    None

    Source:
    None

    Inputs:
    network            - network data structure
    network_face_color - color of panel
    network_edge_color - color of panel edge 

    Properties Used:
    N/A
    """ 
    show_axis     = False 
    save_figure   = False 
    show_figure   = False
    save_filename = 'propulsor'
 
    for propulsor in network.propulsors:   
        number_of_airfoil_points = 21
        tessellation             = 24
        if 'nacelle' in propulsor: 
            if propulsor.nacelle !=  None: 
                plot_data = plot_3d_nacelle(plot_data,propulsor.nacelle,tessellation,number_of_airfoil_points,color_map = 'darkmint') 
        if 'rotor' in propulsor: 
            plot_data = plot_3d_rotor(propulsor.rotor,save_filename,save_figure,plot_data,show_figure,show_axis,0,number_of_airfoil_points,color_map) 
        if 'propeller' in propulsor:
            plot_data = plot_3d_rotor(propulsor.propeller,save_filename,save_figure,plot_data,show_figure,show_axis,0,number_of_airfoil_points,color_map)  
        if 'rotor' in propulsor: 
            plot_data = plot_3d_rotor(propulsor.rotor,save_filename,save_figure,plot_data,show_figure,show_axis,0,number_of_airfoil_points,color_map) 
        if 'propeller' in propulsor:
            plot_data = plot_3d_rotor(propulsor.propeller,save_filename,save_figure,plot_data,show_figure,show_axis,0,number_of_airfoil_points,color_map) 
 
    return plot_data